/*
 * Copyright (C) 2007-2014 Hypertable, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither Hypertable, Inc. nor the names of its contributors may be used to
 *   endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import org.apache.thrift.TException;
import org.apache.thrift.transport.TTransportException;
import org.hypertable.thrift.*;
import org.hypertable.thriftgen.*;

public class HypertableApiTest {

  public static void test_basic(ThriftClient client) throws TException {

    System.out.println("[basic]");

    try {

      if (!client.namespace_exists("test"))
        client.namespace_create("test");

      long ns = client.namespace_open("test");

      boolean if_exists = true;
      client.table_drop(ns, "Fruits", if_exists);

      Schema schema = new Schema();

      Map<String,ColumnFamilySpec> column_families = new HashMap<String,ColumnFamilySpec>();

      ColumnFamilySpec cf = new ColumnFamilySpec();
      cf.setName("genus");
      column_families.put("genus", cf);

      cf = new ColumnFamilySpec();
      cf.setName("description");
      column_families.put("description", cf);

      cf = new ColumnFamilySpec();
      cf.setName("tag");
      column_families.put("tag", cf);

      schema.setColumn_families(column_families);

      client.table_create(ns, "Fruits", schema);

      client.namespace_create("/test/sub");

      List<NamespaceListing> listing;

      listing = client.namespace_get_listing(ns);

      for (NamespaceListing entry : listing)
        System.out.println(entry);

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_convenience(ThriftClient client) throws TException, UnsupportedEncodingException {

    System.out.println("[set cells]");

    try {
      long ns = client.namespace_open("test");

      List<Cell> cells = new ArrayList<Cell>();
      Key key = null;
      Cell cell = null;

      key = new Key();
      key.setRow("apple");
      key.setColumn_family("genus");
      cell = new Cell();
      cell.setKey(key);
      cell.setValue("Malus".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key();
      key.setRow("apple");
      key.setColumn_family("description");
      cell = new Cell();
      cell.setKey(key);
      cell.setValue("The apple is the pomaceous fruit of the apple tree.".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key();
      key.setRow("apple");
      key.setColumn_family("tag");
      key.setColumn_qualifier("crunchy");
      cell = new Cell();
      cell.setKey(key);
      cells.add(cell);

      client.set_cells(ns, "Fruits", cells);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[get cells]");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();
      List<String> columns = new ArrayList<String>();
      columns.add("description");
      ss.setColumns(columns);

      List<Cell> cells = client.get_cells(ns, "Fruits", ss);

      for (Cell cell : cells) {
        String value = null;
        if (cell.value != null)
          value = new String(cell.value.array(), cell.value.position(), cell.value.remaining(), "UTF-8");
        System.out.println(cell.key + " value=" + value);
      }
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[set cells as arrays]");

    try {
      long ns = client.namespace_open("test");
      List<List<String>> cells_as_arrays = new ArrayList<List<String>>();
      List<String> cell_as_array = null;

      cell_as_array = new ArrayList<String>();
      cell_as_array.add("orange");
      cell_as_array.add("genus");
      cell_as_array.add("");
      cell_as_array.add("Citrus");
      cells_as_arrays.add(cell_as_array);

      cell_as_array = new ArrayList<String>();
      cell_as_array.add("orange");
      cell_as_array.add("description");
      cell_as_array.add("");
      cell_as_array.add("The orange (specifically, the sweet orange) is" +
                        "the fruit of the citrus species Citrus x " +
                        "sinensis in the family Rutaceae.");
      cells_as_arrays.add(cell_as_array);

      cell_as_array = new ArrayList<String>();
      cell_as_array.add("orange");
      cell_as_array.add("tag");
      cell_as_array.add("juicy");
      cells_as_arrays.add(cell_as_array);

      client.set_cells_as_arrays(ns, "Fruits", cells_as_arrays);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[get cells as arrays]");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();
      List<String> columns = new ArrayList<String>();
      columns.add("description");
      ss.setColumns(columns);

      List<List<String>> cells_as_arrays = client.get_cells_as_arrays(ns, "Fruits", ss);

      for (List<String> cell_as_array : cells_as_arrays)
        System.out.println(cell_as_array.subList(0, 4));

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[set cells serialized]");

    try {
      long ns = client.namespace_open("test");
      SerializedCellsWriter scw = new SerializedCellsWriter(1024);

      String valueStr = null;
      ByteBuffer valueBuf = null;

      valueStr = "Cucumis";
      valueBuf = ByteBuffer.wrap(valueStr.getBytes("UTF-8"));
      scw.add("canteloupe", "genus", "", SerializedCellsFlag.AUTO_ASSIGN,
              valueBuf);

      valueStr = "Canteloupe refers to a variety of Cucumis melo, a species in "+
        "the family Cucurbitaceae.";
      valueBuf = ByteBuffer.wrap(valueStr.getBytes("UTF-8"));
      scw.add("canteloupe", "description", "", SerializedCellsFlag.AUTO_ASSIGN,
              valueBuf);

      scw.add("canteloupe", "tag", "juicy", SerializedCellsFlag.AUTO_ASSIGN,
              ByteBuffer.allocate(0));

      scw.finalize((byte)0);

      client.set_cells_serialized(ns, "Fruits", scw.buffer());

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[get cells serialized]");

    try {
      long ns = client.namespace_open("test");

      ScanSpec ss = new ScanSpec();

      List<String> columns = new ArrayList<String>();
      columns.add("description");
      ss.setColumns(columns);

      ByteBuffer cells = client.get_cells_serialized(ns, "Fruits", ss);

      SerializedCellsReader reader = new SerializedCellsReader(cells);

      while (reader.next()) {
        String row = new String(reader.get_row(), "UTF-8");
        String column_family = new String(reader.get_column_family(), "UTF-8");
        String column_qualifier = "";
        String value = null;
        if (reader.get_column_qualifier() != null)
          column_qualifier = ":" + new String(reader.get_column_qualifier(), "UTF-8");
        if (reader.get_value() != null)
          value = new String(reader.get_value(), "UTF-8");
        System.out.println(row + "\t" + column_family + column_qualifier + "\t" + value);
      }
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_create_table(ThriftClient client) throws TException {

    System.out.println("[create table]");

    try {
      long ns = client.namespace_open("test");

      Schema schema = new Schema();

      Map<String,AccessGroupSpec> ag_specs = new HashMap<String,AccessGroupSpec>();
      Map<String,ColumnFamilySpec> cf_specs = new HashMap<String,ColumnFamilySpec>();

      // Set table defaults
      {
        AccessGroupOptions ag_options = new AccessGroupOptions();
        ColumnFamilyOptions cf_options = new ColumnFamilyOptions();
        ag_options.setBlocksize(65536);
        schema.setAccess_group_defaults(ag_options);
        cf_options.setMax_versions(1);
        schema.setColumn_family_defaults(cf_options);
      }

      // Access group "ag_normal"
      {
        AccessGroupSpec ag = new AccessGroupSpec();
        ColumnFamilyOptions cf_options = new ColumnFamilyOptions();
        cf_options.setMax_versions(2);
        ag.setName("ag_normal");
        ag.setDefaults(cf_options);
        ag_specs.put("ag_normal", ag);
      }

      // Column "a"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        cf.setName("a");
        cf.setAccess_group("ag_normal");
        cf.setValue_index(true);
        cf.setQualifier_index(true);
        cf_specs.put("a", cf);
      }

      // Column "b"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        ColumnFamilyOptions cf_options = new ColumnFamilyOptions();
        cf.setName("b");
        cf.setAccess_group("ag_normal");
        cf_options.setMax_versions(3);
        cf.setOptions(cf_options);
        cf_specs.put("b", cf);
      }

      // Access group "ag_fast"
      {
        AccessGroupSpec ag = new AccessGroupSpec();
        AccessGroupOptions ag_options = new AccessGroupOptions();
        ag.setName("ag_fast");
        ag_options.setIn_memory(true);
        ag_options.setBlocksize(131072);
        ag.setOptions(ag_options);
        ag_specs.put("ag_fast", ag);
      }

      // Column "c"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        cf.setName("c");
        cf.setAccess_group("ag_fast");
        cf_specs.put("c", cf);
      }

      // Access group "ag_secure"
      {
        AccessGroupSpec ag = new AccessGroupSpec();
        AccessGroupOptions ag_options = new AccessGroupOptions();
        ag.setName("ag_secure");
        ag_options.setReplication((short)5);
        ag.setOptions(ag_options);
        ag_specs.put("ag_secure", ag);
      }

      // Column "d"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        cf.setName("d");
        cf.setAccess_group("ag_secure");
        cf_specs.put("d", cf);
      }

      // Access group "ag_counter"
      {
        AccessGroupSpec ag = new AccessGroupSpec();
        ColumnFamilyOptions cf_options = new ColumnFamilyOptions();
        ag.setName("ag_counter");
        cf_options.setCounter(true);
        cf_options.setMax_versions(0);
        ag.setDefaults(cf_options);
        ag_specs.put("ag_counter", ag);
      }

      // Column "e"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        cf.setName("e");
        cf.setAccess_group("ag_counter");
        cf_specs.put("e", cf);
      }

      // Column "f"
      {
        ColumnFamilySpec cf = new ColumnFamilySpec();
        ColumnFamilyOptions cf_options = new ColumnFamilyOptions();
        cf.setName("f");
        cf.setAccess_group("ag_counter");
        cf_options.setCounter(false);
        cf.setOptions(cf_options);
        cf_specs.put("f", cf);
      }

      schema.setAccess_groups(ag_specs);
      schema.setColumn_families(cf_specs);

      client.table_create(ns, "TestTable", schema);

      HqlResult result = client.hql_query(ns, "SHOW CREATE TABLE TestTable");

      if (!result.results.isEmpty())
        System.out.println(result.results.get(0));

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_alter_table(ThriftClient client) throws TException {

    System.out.println("[alter table]");

    try {
      long ns = client.namespace_open("test");

      Schema schema = client.get_schema(ns, "TestTable");

      // Rename column "b" to "z"
      {
        ColumnFamilySpec cf_spec = schema.column_families.get("b");
        assert cf_spec != null;
        schema.column_families.remove("b");
        cf_spec.setName("z");
        schema.column_families.put("z", cf_spec);
      }

      // Add column "g"
      {
        ColumnFamilySpec cf_spec = new ColumnFamilySpec();
        cf_spec.setName("g");
        cf_spec.setAccess_group("ag_counter");
        schema.column_families.put("g", cf_spec);
      }

      client.table_alter(ns, "TestTable", schema);

      HqlResult result = client.hql_query(ns, "SHOW CREATE TABLE TestTable");

      if (!result.results.isEmpty())
        System.out.println(result.results.get(0));

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static long datetime_to_ns_since_epoch(String dateStr) throws ParseException {
    DateFormat format = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss");
    Date datem = format.parse(dateStr);
    return datem.getTime() * 1000000L;
  }

  public static void test_mutator(ThriftClient client) 
    throws TException, ParseException, UnsupportedEncodingException {

    System.out.println("[mutator]");

    try {
      long ns = client.namespace_open("test");

      List<Cell> cells = new ArrayList<Cell>();
      Key key = null;
      Cell cell = null;
      String valueStr = null;

      long mutator = client.mutator_open(ns, "Fruits", 0, 0);

      // Auto-assigned timestamps

      key = new Key("lemon", "genus", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("Citrus".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("lemon", "tag", "bitter", KeyFlag.INSERT);
      cell = new Cell(key);
      cells.add(cell);

      key = new Key("lemon", "description", null, KeyFlag.INSERT);
      cell = new Cell(key);
      valueStr = "The lemon (Citrus × limon) is a small evergreen tree native" +
        " to Asia.";
      cell.setValue(valueStr.getBytes("UTF-8"));
      cells.add(cell);

      client.mutator_set_cells(mutator, cells);
      client.mutator_flush(mutator);

      // Explicitly-supplied timestamps

      cells.clear();

      key = new Key("mango", "genus", null, KeyFlag.INSERT);
      key.setTimestamp(datetime_to_ns_since_epoch("2014-06-06 16:27:15"));
      cell = new Cell(key);
      cell.setValue("Mangifera".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("mango", "tag", "sweet", KeyFlag.INSERT);
      key.setTimestamp(datetime_to_ns_since_epoch("2014-06-06 16:27:15"));
      cell = new Cell(key);
      cells.add(cell);

      key = new Key("mango", "description", null, KeyFlag.INSERT);
      key.setTimestamp(datetime_to_ns_since_epoch("2014-06-06 16:27:15"));
      cell = new Cell(key);
      valueStr = "Mango is one of the delicious seasonal fruits grown in the" +
        " tropics.";
      cell.setValue(valueStr.getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("mango", "description", null, KeyFlag.INSERT);
      key.setTimestamp(datetime_to_ns_since_epoch("2014-06-06 16:27:16"));
      cell = new Cell(key);
      valueStr = "The mango is a juicy stone fruit belonging to the genus " +
        "Mangifera, consisting of numerous tropical fruiting trees, that are" +
        " cultivated mostly for edible fruits.";
      cell.setValue(valueStr.getBytes("UTF-8"));
      cells.add(cell);

      client.mutator_set_cells(mutator, cells);
      client.mutator_flush(mutator);
  
      // Delete cells

      cells.clear();

      key = new Key("apple", null, null, KeyFlag.DELETE_ROW);
      cell = new Cell(key);
      cells.add(cell);

      key = new Key("mango", "description", null, KeyFlag.DELETE_CELL);
      key.setTimestamp(datetime_to_ns_since_epoch("2014-06-06 16:27:15"));
      cell = new Cell(key);
      cells.add(cell);

      client.mutator_set_cells(mutator, cells);
      client.mutator_flush(mutator);
      client.mutator_close(mutator);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_scanner(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[scanner - full table scan]");

    try {
      long ns = client.namespace_open("test");

      System.out.println("[scanner - full table scan]");

      long scanner = client.scanner_open(ns, "Fruits", new ScanSpec());
      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      };
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[scanner - restricted scan]");

    try {
      long ns = client.namespace_open("test");

      ScanSpec ss = new ScanSpec();
      List<RowInterval> row_intervals = new ArrayList<RowInterval>();

      // Row range [lemon..orange)
      RowInterval ri = new RowInterval();
      ri.setStart_row("lemon");
      ri.setStart_inclusive(true);
      ri.setEnd_row("orange");
      ri.setEnd_inclusive(false);
      row_intervals.add(ri);
      ss.setRow_intervals(row_intervals);

      // Return columns "genus", "tag:bitter", "tag:sweet"
      List<String> columns = new ArrayList<String>();
      columns.add("genus");
      columns.add("tag:bitter");
      columns.add("tag:sweet");
      ss.setColumns(columns);

      // Return only most recent version of each cell
      ss.setVersions(1);
  
      long scanner = client.scanner_open(ns, "Fruits", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      };
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_hql(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[hql query]");

    try {
      long ns = client.namespace_open("test");

      HqlResult result = client.hql_query(ns, "get listing");
      for (String entry : result.results)
        System.out.println(entry);

      result = client.hql_query(ns, "SELECT * from Fruits WHERE ROW = 'pear'");
      for (Cell cell : result.cells) {
        String value = null;
        if (cell.value != null)
          value = new String(cell.value.array(), cell.value.position(),
                             cell.value.remaining(), "UTF-8");
        System.out.println(cell.key + " value=" + value);
      }

      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[HQL query cells as arrays]");

    try {
      long ns = client.namespace_open("test");
      HqlResultAsArrays result_as_arrays =
        client.hql_query_as_arrays(ns, "SELECT * from Fruits WHERE ROW = 'lemon'");
      for (List<String> cell_as_array : result_as_arrays.cells)
        System.out.println(cell_as_array.subList(0, 4));
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[HQL exec - mutator]");

    try {
      long ns = client.namespace_open("test");
      HqlResult result =
        client.hql_exec(ns, "INSERT INTO Fruits VALUES ('strawberry'," +
                        " 'genus', 'Fragaria'), ('strawberry', 'tag:fibrous'," +
                        " ''), ('strawberry', 'description', 'The garden " +
                        "strawberry is a widely grown hybrid species of the " +
                        "genus Fragaria')", true, false);
      List<Cell> cells = new ArrayList<Cell>();
      Cell cell = null;
      Key key = null;

      key = new Key("pineapple", "genus", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("Ananas".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("pineapple", "tag", "acidic", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("Ananas".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("pineapple", "description", null, KeyFlag.INSERT);
      cell = new Cell(key);
      String valueStr =
        "The pineapple (Ananas comosus) is a tropical plant with " +
        "edible multiple fruit consisting of coalesced berries.";
      cell.setValue(valueStr.getBytes("UTF-8"));
      cells.add(cell);

      client.mutator_set_cells(result.mutator, cells);
      client.mutator_flush(result.mutator);
      client.mutator_close(result.mutator);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[HQL exec - scanner]");

    try {
      long ns = client.namespace_open("test");
      HqlResult result = client.hql_exec(ns, "SELECT * from Fruits", false, true);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(result.scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(result.scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_secondary_indices(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[secondary indices]");

    try {
      long ns = client.namespace_open("test");
      String command =
        "CREATE TABLE products (title, section, info, category, INDEX section,"+
        " INDEX info, QUALIFIER INDEX info, QUALIFIER INDEX category)";
      HqlResult result = client.hql_query(ns, command);
      command =
        "LOAD DATA INFILE 'indices_test_products.tsv' INTO TABLE products";
      result = client.hql_query(ns, command);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE section = 'books']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("section");
      column_predicate.setOperation(ColumnPredicateOperation.EXACT_MATCH.getValue());
      column_predicate.setValue("books");
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE info:actor = 'Jack Nicholson']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("actor");
      int operation = ColumnPredicateOperation.EXACT_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Jack Nicholson");
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title,info:publisher " +
                       "FROM products WHERE info:publisher =^ 'Addison-Wesley']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("publisher");
      int operation = ColumnPredicateOperation.PREFIX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Addison-Wesley");
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      columns.add("info:publisher");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title,info:publisher " +
                       "FROM products WHERE info:publisher =~ '^Addison-Wesley']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("publisher");
      int operation = ColumnPredicateOperation.REGEX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("^Addison-Wesley");
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      columns.add("info:publisher");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE Exists(info:studio)]");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("studio");
      column_predicate.setOperation(ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue());
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE Exists(category:/^\\/Movies/)]");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();
      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("category");
      column_predicate.setColumn_qualifier("^/Movies");
      column_predicate.setOperation(ColumnPredicateOperation.QUALIFIER_REGEX_MATCH.getValue());
      column_predicates.add(column_predicate);
      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE info:author =~ /^Stephen P/ OR " +
                       "info:publisher =^ 'Anchor']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();

      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("author");
      int operation = ColumnPredicateOperation.REGEX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("^Stephen P");
      column_predicates.add(column_predicate);

      column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("publisher");
      operation = ColumnPredicateOperation.PREFIX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Anchor");
      column_predicates.add(column_predicate);

      ss.setColumn_predicates(column_predicates);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE info:author =~ /^Stephen [PK]/ AND " +
                       "info:publisher =^ 'Anchor']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();

      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("author");
      int operation = ColumnPredicateOperation.REGEX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("^Stephen [PK]");
      column_predicates.add(column_predicate);

      column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("publisher");
      operation = ColumnPredicateOperation.PREFIX_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Anchor");
      column_predicates.add(column_predicate);

      ss.setColumn_predicates(column_predicates);

      // AND predicates together
      ss.setAnd_column_predicates(true);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products " +
                       "WHERE ROW > 'B00002VWE0' AND " +
                       "info:actor = 'Jack Nicholson']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<RowInterval> row_intervals = new ArrayList<RowInterval>();
      RowInterval ri = new RowInterval();;
      ri.setStart_row("B00002VWE0");
      ri.setStart_inclusive(false);
      row_intervals.add(ri);
      ss.setRow_intervals(row_intervals);

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();

      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("actor");
      int operation = ColumnPredicateOperation.EXACT_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Jack Nicholson");
      column_predicates.add(column_predicate);

      ss.setColumn_predicates(column_predicates);

      ss.setAnd_column_predicates(true);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[secondary index - SELECT title FROM products WHERE " +
                       "ROW =^ 'B' AND " +
                       "info:actor = 'Jack Nicholson']");

    try {
      long ns = client.namespace_open("test");
      ScanSpec ss = new ScanSpec();

      List<RowInterval> row_intervals = new ArrayList<RowInterval>();
      RowInterval ri = new RowInterval();;
      ri.setStart_row("B");
      ri.setStart_inclusive(true);
      ri.setEnd_row("C");
      ri.setEnd_inclusive(false);
      row_intervals.add(ri);
      ss.setRow_intervals(row_intervals);

      List<ColumnPredicate> column_predicates = new ArrayList<ColumnPredicate>();

      ColumnPredicate column_predicate = new ColumnPredicate();
      column_predicate.setColumn_family("info");
      column_predicate.setColumn_qualifier("actor");
      int operation = ColumnPredicateOperation.EXACT_MATCH.getValue() |
        ColumnPredicateOperation.QUALIFIER_EXACT_MATCH.getValue();
      column_predicate.setOperation(operation);
      column_predicate.setValue("Jack Nicholson");
      column_predicates.add(column_predicate);

      ss.setColumn_predicates(column_predicates);

      ss.setAnd_column_predicates(true);

      List<String> columns = new ArrayList<String>();
      columns.add("title");
      ss.setColumns(columns);

      long scanner = client.scanner_open(ns, "products", ss);

      while (true) {
        List<Cell> cells = client.scanner_get_cells(scanner);
        if (cells.isEmpty())
          break;
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_async(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[async]");

    try {
      String command;
      HqlResult result;
      long ns = client.namespace_open("test");
      command = "CREATE TABLE Profile (info, last_access MAX_VERSIONS 1)";
      result = client.hql_query(ns, command);
      command = "CREATE TABLE Session (user_id, page_hit)";
      result = client.hql_query(ns, command);
      command = "INSERT INTO Profile VALUES ('1', 'info:name', 'Joe')," +
        " ('2', 'info:name', 'Sue')";
      result = client.hql_query(ns, command);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[async mutator]");

    try {
      long ns = client.namespace_open("test");
      long ff = client.future_open(0);
      long profile_mutator = client.async_mutator_open(ns, "Profile", ff, 0);
      long session_mutator = client.async_mutator_open(ns, "Session", ff, 0);
      List<Cell> cells = new ArrayList<Cell>();
      Key key;
      Cell cell;

      key = new Key("1", "last_access", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("2014-06-13 16:06:09".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("2", "last_access", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("2014-06-13 16:06:10".getBytes("UTF-8"));
      cells.add(cell);

      client.async_mutator_set_cells(profile_mutator, cells);

      cells.clear();

      key = new Key("0001-200238", "user_id", "1", KeyFlag.INSERT);
      cell = new Cell(key);
      cells.add(cell);

      key = new Key("0001-200238", "page_hit", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("/index.html".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("0002-383049", "user_id", "2", KeyFlag.INSERT);
      cell = new Cell(key);
      cells.add(cell);

      key = new Key("0002-383049", "page_hit", null, KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("/foo/bar.html".getBytes("UTF-8"));
      cells.add(cell);

      client.async_mutator_set_cells(session_mutator, cells);

      client.async_mutator_flush(profile_mutator);
      client.async_mutator_flush(session_mutator);

      int result_count = 0;
      while (true) {
        Result result = client.future_get_result(ff, 0);
        if (result.is_empty)
          break;
        result_count++;
        if (result.is_error) {
          System.out.println("Async mutator error:  " + result.error_msg);
          System.exit(1);
        }
        if (result.id == profile_mutator)
          System.out.println("Result is from Profile mutation");
        else if (result.id == session_mutator)
          System.out.println("Result is from Session mutation");
      }

      System.out.println("result count = " + result_count);

      client.async_mutator_close(profile_mutator);
      client.async_mutator_close(session_mutator);
      client.future_close(ff);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[async scanner]");

    try {
      long ns = client.namespace_open("test");
      long profile_scanner;
      long session_scanner;
      long ff = client.future_open(0);

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("1");
        ri.setStart_inclusive(true);
        ri.setEnd_row("1");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
      }

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("0001-200238");
        ri.setStart_inclusive(true);
        ri.setEnd_row("0001-200238");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
      }

      while (true) {

        Result result = client.future_get_result(ff, 0);

        if (result.is_empty)
          break;

        if (result.is_error) {
          System.out.println("Async scanner error:  " + result.error_msg);
          System.exit(1);
        }

        assert result.is_scan;
        assert result.id == profile_scanner || result.id == session_scanner;

        if (result.id == profile_scanner)
          System.out.println("Result is from Profile scan");
        else if (result.id == session_scanner)
          System.out.println("Result is from Session scan");

        for (Cell cell : result.cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      }

      client.async_scanner_close(profile_scanner);
      client.async_scanner_close(session_scanner);
      client.future_close(ff);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[async scanner - result serialized]");

    try {
      long profile_scanner;
      long session_scanner;
      long ns = client.namespace_open("test");
      long ff = client.future_open(0);

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("1");
        ri.setStart_inclusive(true);
        ri.setEnd_row("1");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
      }

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("0001-200238");
        ri.setStart_inclusive(true);
        ri.setEnd_row("0001-200238");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
      }

      while (true) {

        ResultSerialized result_serialized =
          client.future_get_result_serialized(ff, 0);

        if (result_serialized.is_empty)
          break;

        if (result_serialized.is_error) {
          System.out.println("Async scanner error:  " + result_serialized.error_msg);
          System.exit(1);
        }

        assert(result_serialized.is_scan);
        assert(result_serialized.id == profile_scanner || result_serialized.id == session_scanner);

        if (result_serialized.id == profile_scanner)
          System.out.println("Result is from Profile scan");
        else if (result_serialized.id == session_scanner)
          System.out.println("Result is from Session scan");

        SerializedCellsReader reader =
          new SerializedCellsReader(result_serialized.cells);

        while (reader.next()) {
          String row = new String(reader.get_row(), "UTF-8");
          String column_family = new String(reader.get_column_family(), "UTF-8");
          String column_qualifier = "";
          String value = null;
          if (reader.get_column_qualifier() != null)
            column_qualifier = ":" + new String(reader.get_column_qualifier(), "UTF-8");
          if (reader.get_value() != null)
            value = new String(reader.get_value(), "UTF-8");
          System.out.println(row + "\t" + column_family + column_qualifier + "\t" + value);
        }

      }
      client.async_scanner_close(profile_scanner);
      client.async_scanner_close(session_scanner);
      client.future_close(ff);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[async scanner - result as arrays]");

    try {
      long profile_scanner;
      long session_scanner;
      long ns = client.namespace_open("test");
      long ff = client.future_open(0);

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("1");
        ri.setStart_inclusive(true);
        ri.setEnd_row("1");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
      }

      {
        ScanSpec ss = new ScanSpec();
        List<RowInterval> row_intervals = new ArrayList<RowInterval>();
        RowInterval ri = new RowInterval();
        ri.setStart_row("0001-200238");
        ri.setStart_inclusive(true);
        ri.setEnd_row("0001-200238");
        ri.setEnd_inclusive(true);
        row_intervals.add(ri);
        ss.setRow_intervals(row_intervals);
        session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
      }

      while (true) {

        ResultAsArrays result_as_arrays =
          client.future_get_result_as_arrays(ff, 0);

        if (result_as_arrays.is_empty)
          break;

        if (result_as_arrays.is_error) {
          System.out.println("Async scanner error:  " + result_as_arrays.error_msg);
          System.exit(1);
        }

        assert result_as_arrays.is_scan;
        assert result_as_arrays.id == profile_scanner ||
          result_as_arrays.id == session_scanner;

        if (result_as_arrays.id == profile_scanner)
          System.out.println("Result is from Profile scan");
        else if (result_as_arrays.id == session_scanner)
          System.out.println("Result is from Session scan");

        for (List<String> cell_as_array : result_as_arrays.cells)
          System.out.println(cell_as_array.subList(0, 4));
      }

      client.async_scanner_close(profile_scanner);
      client.async_scanner_close(session_scanner);
      client.future_close(ff);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_atomic_counter(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[atomic counter]");

    try {
      long ns = client.namespace_open("test");
      HqlResult result = client.hql_query(ns, "CREATE TABLE Hits (count COUNTER)");
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[atomic counter - increment]");

    try {
      long ns = client.namespace_open("test");
      long mutator = client.mutator_open(ns, "Hits", 0, 0);

      Key key;
      Cell cell;
      List<Cell> cells = new ArrayList<Cell>();

      key = new Key("/index.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("1".getBytes("UTF-8"));
      cells.add(cell);

      client.mutator_set_cells(mutator, cells);
      client.mutator_flush(mutator);

      long scanner = client.scanner_open(ns, "Hits", new ScanSpec());

      do {
        cells = client.scanner_get_cells(scanner);
        for (Cell cell2 : cells) {
          String value = null;
          if (cell2.value != null)
            value = new String(cell2.value.array(), cell2.value.position(),
                               cell2.value.remaining(), "UTF-8");
          System.out.println(cell2.key + " value=" + value);
        }
      } while (!cells.isEmpty());

      client.scanner_close(scanner);
      client.mutator_close(mutator);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    System.out.println("[atomic counter - reset and subtraction]");

    try {
      long ns = client.namespace_open("test");
      long mutator = client.mutator_open(ns, "Hits", 0, 0);

      Key key;
      Cell cell;
      List<Cell> cells = new ArrayList<Cell>();

      key = new Key("/index.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("=0".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("7".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:18", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("-1".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/index.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("-2".getBytes("UTF-8"));
      cells.add(cell);

      key = new Key("/foo/bar.html", "count", "2014-06-14 07:31:19", KeyFlag.INSERT);
      cell = new Cell(key);
      cell.setValue("=19".getBytes("UTF-8"));
      cells.add(cell);

      client.mutator_set_cells(mutator, cells);
      client.mutator_flush(mutator);

      long scanner = client.scanner_open(ns, "Hits", new ScanSpec());

      do {
        cells = client.scanner_get_cells(scanner);
        for (Cell cell2 : cells) {
          String value = null;
          if (cell2.value != null)
            value = new String(cell2.value.array(), cell2.value.position(),
                               cell2.value.remaining(), "UTF-8");
          System.out.println(cell2.key + " value=" + value);
        }
      } while (!cells.isEmpty());

      client.scanner_close(scanner);
      client.mutator_close(mutator);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void test_unique(ThriftClient client)
    throws TException, UnsupportedEncodingException {

    System.out.println("[unique]");

    try {
      long ns = client.namespace_open("test");
      HqlResult result = client.hql_query(ns, "CREATE TABLE User (info, " +
                                          "id TIME_ORDER desc MAX_VERSIONS 1)");
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    try {
      Key key;
      String ret;
      long ns = client.namespace_open("test");
      key = new Key();
      key.setColumn_family("id");
      key.setRow("joe1987");
      ret = client.create_cell_unique(ns, "User", key, null);
      key = new Key();
      key.setColumn_family("id");
      key.setRow("mary.bellweather");
      ret = client.create_cell_unique(ns, "User", key, null);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }

    Key key = new Key();
    try {
      long ns = client.namespace_open("test");
      key = new Key();
      key.setColumn_family("id");
      key.setRow("joe1987");
      String ret = client.create_cell_unique(ns, "User", key, null);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      if (e.code == org.hypertable.Common.Error.ALREADY_EXISTS)
        System.out.println("User name '" + key.row + "' is already taken");
      else {
        System.out.println(e.message);
        System.exit(1);
      }
    }

    try {
      long ns = client.namespace_open("test");
      long scanner = client.scanner_open(ns, "User", new ScanSpec());
      List<Cell> cells;
      do {
        cells = client.scanner_get_cells(scanner);
        for (Cell cell : cells) {
          String value = null;
          if (cell.value != null)
            value = new String(cell.value.array(), cell.value.position(),
                               cell.value.remaining(), "UTF-8");
          System.out.println(cell.key + " value=" + value);
        }
      } while (!cells.isEmpty());
      client.scanner_close(scanner);
      client.namespace_close(ns);
    }
    catch (ClientException e) {
      System.out.println(e.message);
      System.exit(1);
    }
  }

  public static void main(String args[]) {
    ThriftClient client = null;
    try {
      client = ThriftClient.create("localhost", 15867);
    }
    catch (TException e) {
      System.out.println(e);
      System.exit(1);
    }

    try {
      test_basic(client);
      test_convenience(client);
      test_create_table(client);
      test_alter_table(client);
      test_mutator(client);
      test_scanner(client);
      test_hql(client);
      test_secondary_indices(client);
      test_async(client);
      test_atomic_counter(client);
      test_unique(client);
    }
    catch (TException e) {
      System.out.println(e);
      System.exit(1);
    }
    catch (Exception e) {
      e.printStackTrace();
      System.exit(1);
    }
  }

};
