#!/usr/bin/env perl
#
# Copyright (C) 2007-2014 Hypertable, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither Hypertable, Inc. nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
use Hypertable::ThriftClient;
use Error qw(:try);
use Time::Local;
use strict;
use warnings;

sub print_cell {
    my($cell) = @_;
    my $key = $cell->key;
    print "{Cell key={Key row=" . $key->row . " column_family=" .
        $key->column_family . " column_qualifier=" . $key->column_qualifier .
        "} value=";
    if (defined $cell->value) {
        print $cell->value . "\n";
    }
    else {
        print "[NULL]\n";
    }
}

sub print_cell_as_array {
    my(@cell_as_array) = @_;
    print "{CellAsArray key={Key row=" . $cell_as_array[0] . " column_family=" .
        $cell_as_array[1] . " column_qualifier=" . $cell_as_array[2] .
        "} value=" . $cell_as_array[3] . "\n";
}

sub test_basic {
    my($client) = @_;

    print "[basic]\n";

    try {

        if (!$client->namespace_exists("test")) {
            $client->namespace_create("test");
        }

        my $ns = $client->namespace_open("test");

        my $if_exists = 1;

        $client->table_drop($ns, "Fruits", $if_exists);

        my %cf_specs;
        my $cf_spec;
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "genus"});
        $cf_specs{'genus'} = $cf_spec;
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "description"});
        $cf_specs{'description'} = $cf_spec;
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "tag"});
        $cf_specs{'tag'} = $cf_spec;

        my $schema = new Hypertable::ThriftGen::Schema({column_families => \%cf_specs});

        $client->table_create($ns, "Fruits", $schema);

        $client->namespace_create("/test/sub");

        my $result = $client->namespace_get_listing($ns);

        for my $entry (@{$result}) {
            print $entry->name;
            if ($entry->is_namespace) {
                print "\t(dir)";
            }
            print "\n";
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_convenience {
    my($client) = @_;

    print "[set cells]\n";

    try {
        my $ns = $client->namespace_open("test");

        my @cells;
        my ($value, $key, $cell);

        $key = new Hypertable::ThriftGen::Key({row => 'apple',
                                               column_family => 'genus'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => 'Malus'});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => 'apple',
                                               column_family => 'description'});

        $value = 'The apple is the pomaceous fruit of the apple tree.';
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => $value});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => 'apple',
                                               column_family => 'tag',
                                               column_qualifier => 'cruncy'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        
        $client->set_cells($ns, "Fruits", \@cells);

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[get cells]\n";

    try {
        my $ns = $client->namespace_open("test");

        my @columns;
        push(@columns, "description");
        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns});

        my $result = $client->get_cells($ns, "Fruits", $ss);

        for my $cell (@{$result}) {
            print_cell($cell);
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[set cells as arrays]\n";

    try {
        my $ns = $client->namespace_open("test");      

        my @cells_as_arrays;
        my $cell_as_array;
        my $value;

        $cell_as_array = ["orange", "genus", "", "Citrus"];
        push(@cells_as_arrays, $cell_as_array);

        $value = "The orange (specifically, the sweet orange) is the fruit of " .
            "the citrus species Citrus × sinensis in the family Rutaceae.";
        $cell_as_array = ["orange", "description", "", $value];
        push(@cells_as_arrays, $cell_as_array);

        $cell_as_array = ["orange", "tag", "juicy", ""];
        push(@cells_as_arrays, $cell_as_array);

        $client->set_cells_as_arrays($ns, "Fruits", \@cells_as_arrays);

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[get cells as arrays]\n";

    try {
        my $ns = $client->namespace_open("test");      

        my @columns;
        push(@columns, "description");
        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns});

        my $result = $client->get_cells_as_arrays($ns, "Fruits", $ss);

        for my $cell_as_array (@{$result}) {
            print_cell_as_array(@{$cell_as_array});
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_create_table {
    my($client) = @_;

    print "[create table]\n";

    try {
        my $ns = $client->namespace_open("test");

        my (%ag_specs, %cf_specs);
        my ($ag_spec, $ag_options, $cf_spec, $cf_options);

        # table defaults        
        my $table_ag_defaults = new Hypertable::ThriftGen::AccessGroupOptions({blocksize => 65536});
        my $table_cf_defaults = new Hypertable::ThriftGen::ColumnFamilyOptions({max_versions => 1});

        # Access group "ag_normal"
        $cf_options = new Hypertable::ThriftGen::ColumnFamilyOptions({max_versions => 2});
        $ag_spec = new Hypertable::ThriftGen::AccessGroupSpec({name => 'ag_normal',
                                                               defaults => $cf_options});
        $ag_specs{'ag_normal'} = $ag_spec;

        # Column "a"
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "a",
                                                                access_group => "ag_normal",
                                                                value_index => 1,
                                                                qualifier_index => 1});
        $cf_specs{'a'} = $cf_spec;

        # Column "b"
        $cf_options = new Hypertable::ThriftGen::ColumnFamilyOptions({max_versions => 3});
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "b",
                                                                access_group => "ag_normal",
                                                                options => $cf_options});
        $cf_specs{'b'} = $cf_spec;

        # Access group "ag_fast"
        $ag_options = new Hypertable::ThriftGen::AccessGroupOptions({in_memory => 1,
                                                                     blocksize => 131072});
        $ag_spec = new Hypertable::ThriftGen::AccessGroupSpec({name => 'ag_fast',
                                                               options => $ag_options});
        $ag_specs{'ag_fast'} = $ag_spec;

        # Column "c"
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "c",
                                                                access_group => "ag_fast"});
        $cf_specs{'c'} = $cf_spec;

        # Access group "ag_secure"
        $ag_options = new Hypertable::ThriftGen::AccessGroupOptions({replication => 5});
        $ag_spec = new Hypertable::ThriftGen::AccessGroupSpec({name => 'ag_secure',
                                                               options => $ag_options});
        $ag_specs{'ag_secure'} = $ag_spec;

        # Column "d"
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "d",
                                                                access_group => "ag_secure"});
        $cf_specs{'d'} = $cf_spec;

        # Access group "ag_counter"
        $cf_options = new Hypertable::ThriftGen::ColumnFamilyOptions({max_versions => 0,
                                                                      counter => 1});
        $ag_spec = new Hypertable::ThriftGen::AccessGroupSpec({name => 'ag_counter',
                                                               defaults => $cf_options});
        $ag_specs{'ag_counter'} = $ag_spec;

        # Column "e"
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "e",
                                                                access_group => "ag_counter"});
        $cf_specs{'e'} = $cf_spec;

        # Column "f"
        $cf_options = new Hypertable::ThriftGen::ColumnFamilyOptions({counter => 0});
        $cf_spec = new Hypertable::ThriftGen::ColumnFamilySpec({name => "f",
                                                                access_group => "ag_counter",
                                                                options => $cf_options});
        $cf_specs{'f'} = $cf_spec;

        my $schema = new Hypertable::ThriftGen::Schema({access_group_defaults => $table_ag_defaults,
                                                        column_family_defaults => $table_cf_defaults,
                                                        access_groups => \%ag_specs,
                                                        column_families => \%cf_specs});

        $client->table_create($ns, "TestTable", $schema);

        my $result = $client->hql_query($ns, "SHOW CREATE TABLE TestTable");

        for my $line (@{$result->results}) {
            print $line;
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_alter_table {
    my($client) = @_;

    print "[alter table]\n";

    try {
        my $ns = $client->namespace_open("test");

        my $schema = $client->get_schema($ns, "TestTable");

        my $cf_spec;

        # Rename column "b" to "z"
        my $column_families = $schema->column_families;
        $cf_spec = $column_families->{'b'};
        delete $column_families->{'b'};
        $cf_spec->{name} = 'z';
        $column_families->{'z'} = $cf_spec;

        # Add column "g"
        $cf_spec =
            new Hypertable::ThriftGen::ColumnFamilySpec({name => "g",
                                                         access_group => "ag_counter"});
        $schema->column_families->{'g'} = $cf_spec;

        $client->table_alter($ns, "TestTable", $schema);

        my $result = $client->hql_query($ns, "SHOW CREATE TABLE TestTable");

        for my $line (@{$result->results}) {
            print $line;
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_mutator {
    my($client) = @_;

    print "[mutator]\n";

    try {
        my $ns = $client->namespace_open("test");

        my $mutator = $client->mutator_open($ns, "Fruits", 0, 0);
        my ($timestamp, $description, $key, $cell);
        my @cells;

        # Auto-assigned timestamps
        $key = new Hypertable::ThriftGen::Key({row => 'lemon',
                                               column_family => 'genus'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => 'Citrus'});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => 'lemon',
                                               column_family => 'tag',
                                               column_qualifier => 'bitter'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => 'lemon',
                                               column_family => 'description'});
        $description = "The lemon (Citrus × limon) is a small evergreen " .
            "tree native to Asia.";
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => $description});
        push(@cells, $cell);

        $client->mutator_set_cells($mutator, \@cells);
        $client->mutator_flush($mutator);

        # Explicitly-supplied timestamps
        @cells = ();
        $timestamp = timelocal(15,27,16,6,5,2014) * 1000000000;
        $key = new Hypertable::ThriftGen::Key({row => 'mango',
                                               column_family => 'genus',
                                               timestamp => $timestamp});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => 'Mangifera'});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => 'mango',
                                               column_family => 'tag',
                                               column_qualifier => 'sweet',
                                               timestamp => $timestamp});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $description = "Mango is one of the delicious seasonal fruits grown " .
            "in the tropics.";
        $key = new Hypertable::ThriftGen::Key({row => 'mango',
                                               column_family => 'description',
                                               timestamp => $timestamp});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => $description});
        push(@cells, $cell);
        $timestamp = timelocal(16,27,16,6,5,2014) * 1000000000;
        $description = "The mango is a juicy stone fruit belonging to the " .
            "genus Mangifera, consisting of numerous tropical fruiting trees," .
            "that are cultivated mostly for edible fruits.";
        $key = new Hypertable::ThriftGen::Key({row => 'mango',
                                               column_family => 'description',
                                               timestamp => $timestamp});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => $description});
        push(@cells, $cell);
        $client->mutator_set_cells($mutator, \@cells);
        $client->mutator_flush($mutator);

        # Delete cells
        @cells = ();
        my $flag = Hypertable::ThriftGen::KeyFlag::DELETE_ROW;
        $key = new Hypertable::ThriftGen::Key({row => 'apple',
                                               flag => $flag});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $flag = Hypertable::ThriftGen::KeyFlag::DELETE_CELL;
        $timestamp = timelocal(15,27,16,6,5,2014) * 1000000000;
        $key = new Hypertable::ThriftGen::Key({row => 'mango',
                                               column_family => 'description',
                                               timestamp => $timestamp,
                                               flag => $flag});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $client->mutator_set_cells($mutator, \@cells);
        $client->mutator_flush($mutator);
        $client->mutator_close($mutator);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_scanner {
    my($client) = @_;

    print "[scanner - full table scan]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $ss = new Hypertable::ThriftGen::ScanSpec();
        my $scanner = $client->scanner_open($ns, "Fruits", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[scanner - restricted scan]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $ri = new Hypertable::ThriftGen::RowInterval({start_row => 'lemon',
                                                         start_inclusive => 1,
                                                         end_row => 'orange',
                                                         end_inclusive => 0});
        my @row_intervals = ( $ri );
        my @columns = ( "genus", "tag:bitter", "tag:sweet" );
        my $ss = new Hypertable::ThriftGen::ScanSpec({row_intervals => \@row_intervals,
                                                      columns => \@columns,
                                                      versions => 1});
        my $scanner = $client->scanner_open($ns, "Fruits", $ss);

        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_hql {
    my($client) = @_;

    print "[hql_query]\n";

    try {
        my $ns = $client->namespace_open("test");

        my $result = $client->hql_query($ns, "GET LISTING");

        for my $line (@{$result->results}) {
            print $line . "\n";
        }

        $result = $client->hql_query($ns, "SELECT * from Fruits WHERE ROW = 'mango'");

        for my $cell (@{$result->cells}) {
            print_cell($cell);
        }

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[hql_query_as_arrays]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $command = "SELECT * from Fruits WHERE ROW = 'lemon'";
        my $result_as_arrays = $client->hql_query_as_arrays($ns, $command);
        for my $cell_as_array (@{$result_as_arrays->cells}) {
            print_cell_as_array(@{$cell_as_array});
        }
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[hql_exec mutator]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $command = "INSERT INTO Fruits VALUES ('strawberry', 'genus', " .
            "'Fragaria'), ('strawberry', 'tag:fibrous', ''), ('strawberry', " .
            "'description', 'The garden strawberry is a widely grown hybrid " .
            "species of the genus Fragaria')";

        my $result = $client->hql_exec($ns, $command, 1, 0);
        my ($description, $key, $cell);
        my @cells;

        $key = new Hypertable::ThriftGen::Key({row => 'pineapple',
                                               column_family => 'genus'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => 'Ananas'});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => 'pineapple',
                                               column_family => 'tag',
                                               column_qualifier => 'acidic'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => 'pineapple',
                                               column_family => 'description'});
        $description = "The pineapple (Ananas comosus) is a tropical plant " .
            "with edible multiple fruit consisting of coalesced berries.";
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => $description});
        push(@cells, $cell);

        $client->mutator_set_cells($result->{mutator}, \@cells);
        $client->mutator_flush($result->{mutator});
        $client->mutator_close($result->{mutator});
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[hql_exec scanner]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $result =
            $client->hql_exec($ns, "SELECT * from Fruits", 0, 1);
        while (1) {
            my $cells = $client->scanner_get_cells($result->{scanner});
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($result->{scanner});
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_secondary_indices {
    my($client) = @_;

    print "[secondary indices]\n";

    try {
        my $ns = $client->namespace_open("test");

        my $command = "CREATE TABLE products (title, section, info, category," .
            "INDEX section, INDEX info, QUALIFIER INDEX info," .
            "QUALIFIER INDEX category)";
        my $result = $client->hql_query($ns, $command);

        $command = "LOAD DATA INFILE 'indices_test_products.tsv' INTO TABLE products";
        $result = $client->hql_query($ns, $command);

        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE section = 'books']\n";

    try {
        my $ns = $client->namespace_open("test");

        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "section",
                                                        operation => $operation,
                                                        value => "books"});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE info:actor = 'Jack Nicholson']\n";

    try {
        my $ns = $client->namespace_open("test");

        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::EXACT_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "actor",
                                                        operation => $operation,
                                                        value => "Jack Nicholson"});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =^ 'Addison-Wesley']\n";

    try {
        my $ns = $client->namespace_open("test");
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::PREFIX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "publisher",
                                                        operation => $operation,
                                                        value => "Addison-Wesley"});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title", "info:publisher" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =~ /^Addison-Wesley/]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::REGEX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "publisher",
                                                        operation => $operation,
                                                        value => "^Addison-Wesley"});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title", "info:publisher" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE Exists(info:studio)]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "studio",
                                                        operation => $operation});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE Exists(category:/^\\/Movies/)]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_REGEX_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "category",
                                                        column_qualifier => "^/Movies",
                                                        operation => $operation});
        my @column_predicates = ( $column_predicate );
        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE info:author =~ /^Stephen P/ OR info:publisher =^ 'Anchor']\n";

    try {
        my $ns = $client->namespace_open("test");
        my @column_predicates;

        # info:author =~ /^Stephen P/
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::REGEX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "author",
                                                        operation => $operation,
                                                        value => "^Stephen P"});
        push (@column_predicates, $column_predicate);

        # info:publisher =^ "Anchor"
        $operation = Hypertable::ThriftGen::ColumnPredicateOperation::PREFIX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "publisher",
                                                        operation => $operation,
                                                        value => "Anchor"});
        push (@column_predicates, $column_predicate);

        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE info:author =~ /^Stephen [PK]/ AND info:publisher =^ 'Anchor']\n";

    try {
        my $ns = $client->namespace_open("test");
        my @column_predicates;

        # info:author =~ /^Stephen [PK]/
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::REGEX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "author",
                                                        operation => $operation,
                                                        value => "^Stephen [PK]"});
        push (@column_predicates, $column_predicate);

        # info:publisher =^ "Anchor"
        $operation = Hypertable::ThriftGen::ColumnPredicateOperation::PREFIX_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "publisher",
                                                        operation => $operation,
                                                        value => "Anchor"});
        push (@column_predicates, $column_predicate);

        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      column_predicates => \@column_predicates,
                                                      and_column_predicates => 1});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE ROW > 'B00002VWE0' AND info:actor = 'Jack Nicholson']\n";

    try {
        my $ns = $client->namespace_open("test");

        # ROW > 'B00002VWE0'
        my $ri = new Hypertable::ThriftGen::RowInterval({start_row => "B00002VWE0",
                                                         start_inclusive => 0});
        my @row_intervals = ( $ri );

        # info:actor = 'Jack Nicholson'
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::EXACT_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "actor",
                                                        operation => $operation,
                                                        value => "Jack Nicholson"});
        my @column_predicates = ( $column_predicate );

        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      row_intervals => \@row_intervals,
                                                      column_predicates => \@column_predicates,
                                                      and_column_predicates => 1});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[secondary index - SELECT title FROM products WHERE ROW =^ 'B' AND info:actor = 'Jack Nicholson']\n";

    try {
        my $ns = $client->namespace_open("test");

        # ROW =^ 'B'
        my $ri = new Hypertable::ThriftGen::RowInterval({start_row => "B",
                                                         start_inclusive => 1,
                                                         end_row => "C",
                                                         end_inclusive => 0});
        my @row_intervals = ( $ri );

        # info:actor = 'Jack Nicholson'
        my $operation = Hypertable::ThriftGen::ColumnPredicateOperation::EXACT_MATCH |
            Hypertable::ThriftGen::ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
        my $column_predicate =
            new Hypertable::ThriftGen::ColumnPredicate({column_family => "info",
                                                        column_qualifier => "actor",
                                                        operation => $operation,
                                                        value => "Jack Nicholson"});
        my @column_predicates = ( $column_predicate );

        my @columns = ( "title" );

        my $ss = new Hypertable::ThriftGen::ScanSpec({columns => \@columns,
                                                      row_intervals => \@row_intervals,
                                                      column_predicates => \@column_predicates,
                                                      and_column_predicates => 1});
        my $scanner = $client->scanner_open($ns, "products", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_async {
    my($client) = @_;

    print "[async]\n";

    try {
        my $ns = $client->namespace_open("test");
        my ($command, $result);
        $command = "CREATE TABLE Profile (info, last_access MAX_VERSIONS 1)";
        $result = $client->hql_query($ns, $command);
        $command = "CREATE TABLE Session (user_id, page_hit)";
        $result = $client->hql_query($ns, $command);
        $command = "INSERT INTO Profile VALUES ('1', 'info:name', 'Joe'), ('2', 'info:name', 'Sue')";
        $result = $client->hql_query($ns, $command);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[async mutator]\n";

    try {
        my $ns = $client->namespace_open("test");

        my $ff = $client->future_open(0);
        my $profile_mutator = $client->async_mutator_open($ns, "Profile", $ff, 0);
        my $session_mutator = $client->async_mutator_open($ns, "Session", $ff, 0);
        my ($key, $cell);
        my @cells;

        $key = new Hypertable::ThriftGen::Key({row => '1',
                                               column_family => 'last_access'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "2014-06-13 16:06:09"});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => '2',
                                               column_family => 'last_access'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "2014-06-13 16:06:10"});
        push(@cells, $cell);
        $client->async_mutator_set_cells($profile_mutator, \@cells);

        @cells = ();
        $key = new Hypertable::ThriftGen::Key({row => "0001-200238",
                                               column_family => 'user_id',
                                               column_qualifier => "1"});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => "0001-200238",
                                               column_family => 'page_hit'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "/index.html"});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => "0002-383049",
                                               column_family => 'user_id',
                                               column_qualifier => "2"});
        $cell = new Hypertable::ThriftGen::Cell({key => $key});
        push(@cells, $cell);
        $key = new Hypertable::ThriftGen::Key({row => "0002-383049",
                                               column_family => 'page_hit'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "/foo/bar.html"});
        push(@cells, $cell);
        $client->async_mutator_set_cells($session_mutator, \@cells);

        $client->async_mutator_flush($profile_mutator);
        $client->async_mutator_flush($session_mutator);

        my $result_count = 0;
        while (1) {
            my $result = $client->future_get_result($ff, 0);
            if ($result->is_empty) {
                last;
            }
            $result_count++;
            if ($result->is_error) {
                print "Async mutator error:  " . $result->error_msg;
                exit 1;
            }
            if ($result->id == $profile_mutator) {
                print "Result is from Profile mutation\n";
            }
            elsif ($result->id == $session_mutator) {
                print "Result is from Session mutation\n";
            }
        }

        print "result count = $result_count\n";
        $client->async_mutator_close($profile_mutator);
        $client->async_mutator_close($session_mutator);
        $client->future_close($ff);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[async scanner]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $ff = $client->future_open(0);


        my $ri = new Hypertable::ThriftGen::RowInterval({start_row => "1",
                                                         start_inclusive => 1,
                                                         end_row => "1",
                                                         end_inclusive => 1});
        my @row_intervals = ( $ri );
        my $ss = new Hypertable::ThriftGen::ScanSpec({row_intervals => \@row_intervals});
        my $profile_scanner = $client->async_scanner_open($ns, "Profile", $ff, $ss);

        $ri = new Hypertable::ThriftGen::RowInterval({start_row => "0001-200238",
                                                         start_inclusive => 1,
                                                         end_row => "0001-200238",
                                                         end_inclusive => 1});
        @row_intervals = ( $ri );
        $ss = new Hypertable::ThriftGen::ScanSpec({row_intervals => \@row_intervals});
        my $session_scanner = $client->async_scanner_open($ns, "Session", $ff, $ss);

        while (1) {
            my $result = $client->future_get_result($ff, 0);
            if ($result->is_empty) {
                last;
            }
            if ($result->is_error) {
                print "Async scanner error:  " . $result->error_msg;
                exit 1;
            }

            $result->is_scan || die "Result is not from scan";
            $result->id == $profile_scanner || $result->id == $session_scanner ||
                die "ID does not match any of the scanners";

            if ($result->id == $profile_scanner) {
                print "Result is from Profile scan\n";
            }
            elsif ($result->id == $session_scanner) {
                print "Result is from Session scan\n";
            }

            for my $cell (@{$result->cells}) {
                print_cell($cell);
            }
        }

        $client->async_scanner_close($profile_scanner);
        $client->async_scanner_close($session_scanner);
        $client->future_close($ff);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[async scanner - result as arrays]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $ff = $client->future_open(0);


        my $ri = new Hypertable::ThriftGen::RowInterval({start_row => "1",
                                                         start_inclusive => 1,
                                                         end_row => "1",
                                                         end_inclusive => 1});
        my @row_intervals = ( $ri );
        my $ss = new Hypertable::ThriftGen::ScanSpec({row_intervals => \@row_intervals});
        my $profile_scanner = $client->async_scanner_open($ns, "Profile", $ff, $ss);

        $ri = new Hypertable::ThriftGen::RowInterval({start_row => "0001-200238",
                                                         start_inclusive => 1,
                                                         end_row => "0001-200238",
                                                         end_inclusive => 1});
        @row_intervals = ( $ri );
        $ss = new Hypertable::ThriftGen::ScanSpec({row_intervals => \@row_intervals});
        my $session_scanner = $client->async_scanner_open($ns, "Session", $ff, $ss);

        while (1) {
            my $result_as_arrays = $client->future_get_result_as_arrays($ff, 0);
            if ($result_as_arrays->is_empty) {
                last;
            }
            if ($result_as_arrays->is_error) {
                print "Async scanner error:  " . $result_as_arrays->error_msg;
                exit 1;
            }

            $result_as_arrays->is_scan || die "Result is not from scan";
            $result_as_arrays->id == $profile_scanner ||
                $result_as_arrays->id == $session_scanner ||
                die "ID does not match any of the scanners";

            if ($result_as_arrays->id == $profile_scanner) {
                print "Result is from Profile scan\n";
            }
            elsif ($result_as_arrays->id == $session_scanner) {
                print "Result is from Session scan\n";
            }

            for my $cell_as_array (@{$result_as_arrays->cells}) {
                print_cell_as_array(@{$cell_as_array});
            }
        }

        $client->async_scanner_close($profile_scanner);
        $client->async_scanner_close($session_scanner);
        $client->future_close($ff);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

}

sub test_atomic_counter {
    my($client) = @_;

    print "[atomic counter]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $result = $client->hql_query($ns, "CREATE TABLE Hits (count COUNTER)");
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[atomic counter - increment]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $mutator = $client->mutator_open($ns, "Hits", 0, 0);

        my ($key, $cell);
        my @cells;

        $key = new Hypertable::ThriftGen::Key({row => '/index.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:18'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/foo/bar.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:18'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/index.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:19'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/foo/bar.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:19'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "1"});
        push(@cells, $cell);

        $client->mutator_set_cells($mutator, \@cells);
        $client->mutator_flush($mutator);

        my $ss = new Hypertable::ThriftGen::ScanSpec();
        my $scanner = $client->scanner_open($ns, "Hits", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->mutator_close($mutator);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    print "[atomic counter - reset and subtraction]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $mutator = $client->mutator_open($ns, "Hits", 0, 0);

        my ($key, $cell);
        my @cells;

        $key = new Hypertable::ThriftGen::Key({row => '/index.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:18'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "=0"});
        push(@cells, $cell);
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "7"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/foo/bar.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:18'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "-1"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/index.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:19'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "-2"});
        push(@cells, $cell);

        $key = new Hypertable::ThriftGen::Key({row => '/foo/bar.html',
                                               column_family => 'count',
                                               column_qualifier => '2014-06-14 07:31:19'});
        $cell = new Hypertable::ThriftGen::Cell({key => $key, value => "=19"});
        push(@cells, $cell);

        $client->mutator_set_cells($mutator, \@cells);
        $client->mutator_flush($mutator);

        my $ss = new Hypertable::ThriftGen::ScanSpec();
        my $scanner = $client->scanner_open($ns, "Hits", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->mutator_close($mutator);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

sub test_unique {
    my($client) = @_;

    print "[unique]\n";

    try {
        my $ns = $client->namespace_open("test");
        my $command = "CREATE TABLE User (info, id TIME_ORDER desc MAX_VERSIONS 1)";
        my $result = $client->hql_query($ns, $command);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };


    my ($ns, $ret, $key);

    try {
        $ns = $client->namespace_open("test");
        $key = new Hypertable::ThriftGen::Key({row => 'joe1987',
                                               column_family => 'id'});
        $ret = $client->create_cell_unique($ns, "User", $key, "");
        $key = new Hypertable::ThriftGen::Key({row => 'mary.bellweather',
                                               column_family => 'id'});
        $ret = $client->create_cell_unique($ns, "User", $key, "");
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };

    try {
        $ns = $client->namespace_open("test");
        $key = new Hypertable::ThriftGen::Key({row => 'joe1987',
                                               column_family => 'id'});
        $ret = $client->create_cell_unique($ns, "User", $key, "");
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        if ($ex->{code} == 48) {
            print "User name '$key->{row}' is already taken\n";
            $client->namespace_close($ns);
        }
        else {
            print "error: $ex->{message}\n";
            exit 1;
        }
    };

    try {
        $ns = $client->namespace_open("test");
        my $ss = new Hypertable::ThriftGen::ScanSpec();
        my $scanner = $client->scanner_open($ns, "User", $ss);
        while (1) {
            my $cells = $client->scanner_get_cells($scanner);
            if (@{$cells}) {
                for my $cell (@{$cells}) {
                    print_cell($cell);
                }
            }
            else {
                last;
            }
        }
        $client->scanner_close($scanner);
        $client->namespace_close($ns);
    }
    catch Hypertable::ThriftGen::ClientException with {
        my $ex = shift;
        print "error: $ex->{message}\n";
        exit 1;
    };
}

my $client = new Hypertable::ThriftClient("localhost", 15867);

&test_basic($client);
&test_convenience($client);
&test_create_table($client);
&test_alter_table($client);
&test_mutator($client);
&test_scanner($client);
&test_hql($client);
&test_secondary_indices($client);
&test_async($client);
&test_atomic_counter($client);
&test_unique($client);

exit 0;
