#
# Copyright (C) 2007-2014 Hypertable, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither Hypertable, Inc. nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import sys
import ctypes
import datetime
import time
import libHyperPython
from hypertable.thriftclient import *
from hyperthrift.gen.ttypes import *

def print_cell_as_array(cell):
  "Prints a cell as array without the timestamp"
  print cell[0:4];

def test_basic(client):

  print "[basic]";

  try:    
    if not client.namespace_exists("test"):
      client.namespace_create("test");
    
    ns = client.namespace_open("test");
    
    if_exists = True;
    client.table_drop(ns, "Fruits", if_exists);
    
    column_families = {};
    column_families["genus"] = ColumnFamilySpec("genus");
    column_families["description"] = ColumnFamilySpec("description");
    column_families["tag"] = ColumnFamilySpec("tag");
    schema = Schema(column_families = column_families);
    client.table_create(ns, "Fruits", schema);
    
    client.namespace_create("/test/sub");
    
    listing = client.namespace_get_listing(ns);
    for entry in listing:
      if entry.is_namespace:
        print "{0}\t(dir)".format(entry.name);
      else:
        print "{0}".format(entry.name);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_convenience(client):

  print "[set_cells]";

  try:
    ns = client.namespace_open("test");
    
    cells = [];
    
    key = Key(row = "apple", column_family = "genus");
    cell = Cell(key, "Malus");
    cells.append(cell);
    
    key = Key(row = "apple", column_family = "description");
    cell = Cell(key, "The apple is the pomaceous fruit of the apple tree.");
    cells.append(cell);
    
    key = Key(row = "apple", column_family = "tag", column_qualifier = "crunchy");
    cell = Cell(key, None);
    cells.append(cell);

    client.set_cells(ns, "Fruits", cells);

    client.namespace_close(ns);

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[get cells]";

  try:
    ns = client.namespace_open("test");
    
    ss = ScanSpec();
    ss.columns = ["description"];
    
    cells = client.get_cells(ns, "Fruits", ss);

    for cell in cells:
      print cell;

    client.namespace_close(ns);

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[set cells as arrays]";

  try:
    ns = client.namespace_open("test");
    
    cells_as_arrays = [];
    
    cell_as_array = ["orange", "genus", "", "Citrus"];
    cells_as_arrays.append(cell_as_array);
    
    cell_as_array = ["orange", "description", "", "The orange (specifically, "
                     "the sweet orange) is the fruit of the citrus species "
                     "Citrus x sinensis in the family Rutaceae."];
    cells_as_arrays.append(cell_as_array);

    cell_as_array = ["orange", "tag", "juicy", ""];
    cells_as_arrays.append(cell_as_array);
    
    client.set_cells_as_arrays(ns, "Fruits", cells_as_arrays);

    client.namespace_close(ns);

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[get cells as arrays]";

  try:
    ns = client.namespace_open("test");

    ss = ScanSpec();
    ss.columns = ["description"];
    
    cells_as_arrays = client.get_cells_as_arrays(ns, "Fruits", ss);
    
    for cell_as_array in cells_as_arrays:
      print_cell_as_array(cell_as_array);

    client.namespace_close(ns);

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[set cells serialized]";

  try:
    ns = client.namespace_open("test");
    
    writer = libHyperPython.SerializedCellsWriter(100, 1)

    writer.add("canteloupe", "genus", "", 0, "Cucumis", 8, KeyFlag.INSERT);
    writer.add("canteloupe", "description", "", 0, "Canteloupe refers to a "
               "variety of Cucumis melo, a species in the family Cucurbitaceae.",
               87, KeyFlag.INSERT);
    writer.add("canteloupe", "tag", "juicy", 0, "", 0, KeyFlag.INSERT);

    writer.finalize(0);

    client.set_cells_serialized(ns, "Fruits", writer.get())

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[get cells serialized]";

  try:
    ns = client.namespace_open("test");

    ss = ScanSpec();
    ss.columns = ["description"];

    cells_serialized = client.get_cells_serialized(ns, "Fruits", ss);
    
    reader = libHyperPython.SerializedCellsReader(cells_serialized, len(cells_serialized))
    while reader.has_next():
      cell = reader.get_cell();
      print cell;
    
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_create_table(client):
  
  print "[create table]";

  try:  
    ns = client.namespace_open("test");

    default_ag_options = AccessGroupOptions(blocksize = 65536);
    default_cf_options = ColumnFamilyOptions(max_versions = 1);
  
    ag_specs = {};
    cf_specs = {};

    cf_options = ColumnFamilyOptions(max_versions = 2);
    ag_spec = AccessGroupSpec(name = "ag_normal", defaults = cf_options);
    ag_specs["ag_normal"] = ag_spec;
  
    cf_spec = ColumnFamilySpec(name = "a", access_group = "ag_normal",
                               value_index = True, qualifier_index = True);
    cf_specs["a"] = cf_spec;

    cf_options = ColumnFamilyOptions(max_versions = 3);
    cf_spec = ColumnFamilySpec(name = "b", access_group = "ag_normal",
                               options = cf_options);
    cf_specs["b"] = cf_spec;

    ag_options = AccessGroupOptions(in_memory = True, blocksize = 131072);
    ag_spec = AccessGroupSpec(name = "ag_fast", options = ag_options);
    ag_specs["ag_fast"] = ag_spec;
  
    cf_spec = ColumnFamilySpec(name = "c", access_group = "ag_fast");
    cf_specs["c"] = cf_spec;

    ag_options = AccessGroupOptions(replication = 5);
    ag_spec = AccessGroupSpec(name = "ag_secure", options = ag_options);
    ag_specs["ag_secure"] = ag_spec;

    cf_spec = ColumnFamilySpec(name = "d", access_group = "ag_secure");
    cf_specs["d"] = cf_spec;

    cf_options = ColumnFamilyOptions(counter = True, max_versions = 0);
    ag_spec = AccessGroupSpec(name = "ag_counter", defaults = cf_options);
    ag_specs["ag_counter"] = ag_spec;

    cf_spec = ColumnFamilySpec(name = "e", access_group = "ag_counter");
    cf_specs["e"] = cf_spec;

    cf_options = ColumnFamilyOptions(counter = False);
    cf_spec = ColumnFamilySpec(name = "f", access_group = "ag_counter",
                               options = cf_options);
    cf_specs["f"] = cf_spec;
  
    schema = Schema(ag_specs, cf_specs, None, None, None,
                    default_ag_options, default_cf_options);
      
    client.table_create(ns, "TestTable", schema);

    result = client.hql_query(ns, "SHOW CREATE TABLE TestTable");

    print result.results[0];

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_alter_table(client):
  
  print "[alter table]";

  try:
    ns = client.namespace_open("test");
    
    schema = client.get_schema(ns, "TestTable");

    # Rename column "b" to "z"
    cf_spec = schema.column_families["b"];
    del schema.column_families["b"];
    cf_spec.name = "z";
    schema.column_families["z"] = cf_spec;

    # Add column "g"
    cf_spec = ColumnFamilySpec(name = "g", access_group = "ag_counter");
    schema.column_families["g"] = cf_spec;

    client.table_alter(ns, "TestTable", schema);
    
    result = client.hql_query(ns, "SHOW CREATE TABLE TestTable");
    
    print result.results[0];
    
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
def test_mutator(client):

  print "[mutator]";

  try:
    ns = client.namespace_open("test");
    
    mutator = client.mutator_open(ns, "Fruits", 0, 0);
    
    ## Auto-assigned timestamps

    cells = [];

    key = Key(row = "lemon", column_family = "genus");
    cell = Cell(key, "Citrus");
    cells.append(cell);
    
    key = Key(row = "lemon", column_family = "tag", column_qualifier = "bitter");
    cell = Cell(key, None);
    cells.append(cell);

    key = Key(row = "lemon", column_family = "description");
    cell = Cell(key, "The lemon (Citrus x limon) is a small evergreen tree "
                "native to Asia.");
    cells.append(cell);

    client.mutator_set_cells(mutator, cells);
    client.mutator_flush(mutator);

    ## Explicitly-supplied timestamps

    cells = [];

    # 2014-06-06 16:27:15
    dt = datetime.datetime(2014, 6, 6, 16, 27, 15);
    ts = long(time.mktime(dt.timetuple())) * 1000000000L;

    key = Key(row = "mango", column_family = "genus", timestamp = ts);
    cell = Cell(key, "Mangifera");
    cells.append(cell);

    key = Key(row = "mango", column_family = "tag", column_qualifier = "sweet",
              timestamp = ts);
    cell = Cell(key, None);
    cells.append(cell);

    key = Key(row = "mango", column_family = "description", timestamp = ts);
    cell = Cell(key, "Mango is one of the delicious seasonal fruits grown in "
                "the tropics.");
    cells.append(cell);

    # 2014-06-06 16:27:16
    dt = datetime.datetime(2014, 6, 6, 16, 27, 16);
    ts = long(time.mktime(dt.timetuple())) * 1000000000L;

    key = Key(row = "mango", column_family = "description", timestamp = ts);
    cell = Cell(key, "The mango is a juicy stone fruit belonging to the genus "
                "Mangifera, consisting of numerous tropical fruiting trees, "
                "that are cultivated mostly for edible fruits.");
    cells.append(cell);

    client.mutator_set_cells(mutator, cells);
    client.mutator_flush(mutator);

    cells = [];

    ## Delete clells

    dt = datetime.datetime(2014, 6, 6, 16, 27, 15);
    ts = long(time.mktime(dt.timetuple())) * 1000000000L;

    key = Key(row = "apple", flag = KeyFlag.DELETE_ROW);
    cell = Cell(key, None);
    cells.append(cell);

    dt = datetime.datetime(2014, 6, 6, 16, 27, 15);
    ts = long(time.mktime(dt.timetuple())) * 1000000000L;
    key = Key(row = "mango", column_family = "description",
              timestamp = ts, flag = KeyFlag.DELETE_CELL);
    cell = Cell(key, None);
    cells.append(cell);
    
    client.mutator_set_cells(mutator, cells);
    client.mutator_flush(mutator);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_scanner(client):

  print "[scanner full table scan]";

  try:
    ns = client.namespace_open("test");

    scanner = client.scanner_open(ns, "Fruits", ScanSpec());

    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[scanner restricted scan]";

  try:
    ns = client.namespace_open("test");

    row_intervals = [ RowInterval("cherry", True, "orange", False) ];
    columns = ["genus", "tag:fleshy", "tag:bitter", "tag:sweet"];
    ss = ScanSpec(row_intervals = row_intervals, versions = 1, columns = columns);
    
    scanner = client.scanner_open(ns, "Fruits", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
    
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_hql(client):

  print "[HQL query]";

  try:
    ns = client.namespace_open("test");
    
    result = client.hql_query(ns, "GET LISTING");
    for line in result.results:
      print line;
    
    result = client.hql_query(ns, "SELECT * from Fruits WHERE ROW = 'mango'");
    for cell in result.cells:
      print cell;

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[HQL query cells as arrays]";

  try:
    ns = client.namespace_open("test");

    result_as_arrays = \
        client.hql_query_as_arrays(ns, "SELECT * from Fruits WHERE ROW = 'lemon'");
    for cell_as_array in result_as_arrays.cells:
      print_cell_as_array(cell_as_array);

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
    
  print "[HQL exec - mutator]";

  try:
    ns = client.namespace_open("test");

    result = \
        client.hql_exec(ns, "INSERT INTO Fruits VALUES ('strawberry', 'genus',"
                        "'Fragaria'), ('strawberry', 'tag:fibrous', ''), "
                        "('strawberry', 'description', 'The garden strawberry "
                        "is a widely grown hybrid species of the genus "
                        "Fragaria')", True, False);
    cells = [];
    
    key = Key(row = "pineapple", column_family = "genus");
    cell = Cell(key, "Anans");
    cells.append(cell);

    key = Key(row = "pineapple", column_family = "tag", column_qualifier = "acidic");
    cell = Cell(key, None);
    cells.append(cell);
        
    key = Key(row = "pineapple", column_family = "description");
    cell = Cell(key, "The pineapple (Ananas comosus) is a tropical plant with "
                "edible multiple fruit consisting of coalesced berries.");
    cells.append(cell);

    client.mutator_set_cells(result.mutator, cells);
    client.mutator_flush(result.mutator);
    client.mutator_close(result.mutator);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[HQL exec - scanner]";

  try:
    ns = client.namespace_open("test");

    result = client.hql_exec(ns, "SELECT * from Fruits", False, True);
    
    while True:
      cells = client.scanner_get_cells(result.scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(result.scanner);
    
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_secondary_indices(client):
  
  print "[secondary indices]";

  try:
    ns = client.namespace_open("test");
    
    command = ("CREATE TABLE products (title, section, info, category, INDEX "
               "section, INDEX info, QUALIFIER INDEX info, QUALIFIER INDEX category)");
    result = client.hql_query(ns, command);

    command = "LOAD DATA INFILE 'indices_test_products.tsv' INTO TABLE products";
    client.hql_query(ns, command);

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print "[secondary index - SELECT title FROM products WHERE section = 'books']";

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("section", None,
                                       ColumnPredicateOperation.EXACT_MATCH,
                                       "books");
    columns = [ "title" ];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print "[secondary index - SELECT title FROM products WHERE info:actor = 'Jack Nicholson'";

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("info", "actor",
                                       ColumnPredicateOperation.EXACT_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Jack Nicholson");
    columns = [ "title" ];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);

    client.namespace_close(ns);
      
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title,info:publisher FROM products WHERE "
         "info:publisher =^ 'Addison-Wesley']");

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("info", "publisher",
                                       ColumnPredicateOperation.PREFIX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Addison-Wesley");
    columns = ["title", "info:publisher"];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title,info:publisher FROM products WHERE "
         "info:publisher =~ '^Addison-Wesley']");

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("info", "publisher",
                                       ColumnPredicateOperation.REGEX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "^Addison-Wesley");
    columns = ["title", "info:publisher"];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  print ("[secondary index - SELECT title FROM products WHERE "
         "Exists(info:studio)]");

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("info", "studio",
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       None);
    columns = ["title"];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title FROM products WHERE "
         "Exists(category:/^\/Movies/)]");

  try:
    ns = client.namespace_open("test");

    column_predicate = ColumnPredicate("category", "^/Movies",
                                       ColumnPredicateOperation.QUALIFIER_REGEX_MATCH,
                                       None);
    columns = ["title"];
    ss = ScanSpec(column_predicates = [column_predicate], columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title FROM products WHERE info:author =~ "
         "/^Stephen P/ OR info:publisher =^ 'Anchor']");

  try:
    ns = client.namespace_open("test");

    predicates = [];
    column_predicate = ColumnPredicate("info", "author",
                                       ColumnPredicateOperation.REGEX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "^Stephen P");
    predicates.append(column_predicate);
    column_predicate = ColumnPredicate("info", "publisher",
                                       ColumnPredicateOperation.PREFIX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Anchor");
    predicates.append(column_predicate);

    columns = ["title"];
    ss = ScanSpec(column_predicates = predicates, columns = columns);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title FROM products WHERE info:author =~ "
         "/^Stephen [PK]/ AND info:publisher =^ 'Anchor']");

  try:
    ns = client.namespace_open("test");

    predicates = [];
    column_predicate = ColumnPredicate("info", "author",
                                       ColumnPredicateOperation.REGEX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "^Stephen [PK]");
    predicates.append(column_predicate);
    column_predicate = ColumnPredicate("info", "publisher",
                                       ColumnPredicateOperation.PREFIX_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Anchor");
    predicates.append(column_predicate);

    columns = ["title"];
    ss = ScanSpec(column_predicates = predicates, columns = columns,
                  and_column_predicates = True);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title FROM products WHERE "
         "ROW > 'B00002VWE0' AND info:actor = 'Jack Nicholson']");

  try:
    ns = client.namespace_open("test");

    row_interval = RowInterval("B00002VWE0", False, None, True);
    column_predicate = ColumnPredicate("info", "actor",
                                       ColumnPredicateOperation.EXACT_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Jack Nicholson");

    ss = ScanSpec(row_intervals = [row_interval],
                  column_predicates = [column_predicate],
                  columns = ["title"], and_column_predicates = True);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print ("[secondary index - SELECT title FROM products WHERE ROW =^ 'B' AND "
         "info:actor = 'Jack Nicholson']");

  try:
    ns = client.namespace_open("test");

    row_interval = RowInterval("B", True, "C", False);
    column_predicate = ColumnPredicate("info", "actor",
                                       ColumnPredicateOperation.EXACT_MATCH |
                                       ColumnPredicateOperation.QUALIFIER_EXACT_MATCH,
                                       "Jack Nicholson");

    ss = ScanSpec(row_intervals = [row_interval],
                  column_predicates = [column_predicate],
                  columns = ["title"], and_column_predicates = True);
    
    scanner = client.scanner_open(ns, "products", ss);
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
  
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_async(client):
  
  print "[async]";
  
  try:
    ns = client.namespace_open("test");
    
    command = "CREATE TABLE Profile (info, last_access MAX_VERSIONS 1)";
    result = client.hql_query(ns, command);

    command = "CREATE TABLE Session (user_id, page_hit)";
    client.hql_query(ns, command);

    command = ("INSERT INTO Profile VALUES ('1', 'info:name', 'Joe'),"
               " ('2', 'info:name', 'Sue')");
    client.hql_query(ns, command);

    client.namespace_close(ns);

  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[async mutator]";

  try:
    ns = client.namespace_open("test");
    ff = client.future_open(0);
    profile_mutator = client.async_mutator_open(ns, "Profile", ff, 0);
    session_mutator = client.async_mutator_open(ns, "Session", ff, 0);

    cells = [];

    key = Key(row = "1", column_family = "last_access");
    cell = Cell(key, "2014-06-13 16:06:09");
    cells.append(cell);

    key = Key(row = "2", column_family = "last_access");
    cell = Cell(key, "2014-06-13 16:06:10");
    cells.append(cell);

    client.async_mutator_set_cells(profile_mutator, cells);
    
    cells = [];
    
    key = Key(row = "0001-200238", column_family = "user_id",
              column_qualifier = "1");
    cell = Cell(key, None);
    cells.append(cell);
    
    key = Key(row = "0001-200238", column_family = "page_hit");
    cell = Cell(key, "/index.html");
    cells.append(cell);
    
    key = Key(row = "0002-383049", column_family = "user_id",
              column_qualifier = "2");
    cell = Cell(key, None);
    cells.append(cell);
    
    key = Key(row = "0002-383049", column_family = "page_hit");
    cell = Cell(key, "/foo/bar.html");
    cells.append(cell);
    
    client.async_mutator_set_cells(session_mutator, cells);
    
    client.async_mutator_flush(profile_mutator);
    client.async_mutator_flush(session_mutator);

    result_count = 0;
    while True:
      result = client.future_get_result(ff, 0);
      if result.is_empty:
        break;
      result_count += 1;
      if result.is_error:
        print "Async mutator error:  {0}".format(result.error_msg);
        sys.exit(1);
      if result.id == profile_mutator:
        print "Result is from Profile mutation";
      elif result.id == session_mutator:
        print "Result is from Session mutation";
    
    print "result count = {0}".format(result_count);
    
    client.async_mutator_close(profile_mutator);
    client.async_mutator_close(session_mutator);
    client.future_close(ff);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print "[async scanner]";

  try:
    ns = client.namespace_open("test");
    ff = client.future_open(0);

    row_intervals = [ RowInterval("1", True, "1", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
    
    row_intervals = [ RowInterval("0001-200238", True, "0001-200238", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
    
    while True:
      result = client.future_get_result(ff, 0);

      if result.is_empty:
        break;

      if result.is_error:
        print "Async scanner error:  {0}".format(result.error_msg);
        sys.exit(1);
      
      assert result.is_scan;
      assert result.id == profile_scanner or result.id == session_scanner;

      if result.id == profile_scanner:
        print "Result is from Profile scan";
      elif result.id == session_scanner:
        print "Result is from Session scan";
      
      for cell in result.cells:
        print cell;

    client.async_scanner_close(profile_scanner);
    client.async_scanner_close(session_scanner);
    client.future_close(ff);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print "[async scanner - result serialized]";

  try:
    ns = client.namespace_open("test");
    ff = client.future_open(0);

    row_intervals = [ RowInterval("1", True, "1", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
    
    row_intervals = [ RowInterval("0001-200238", True, "0001-200238", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
    
    while True:
      result_serialized = client.future_get_result_serialized(ff, 0);

      if result_serialized.is_empty:
        break;

      if result_serialized.is_error:
        print "Async scanner error:  {0}".format(result_serialized.error_msg);
        sys.exit(1);
      
      assert result_serialized.is_scan;
      assert result_serialized.id == profile_scanner or result_serialized.id == session_scanner;

      if result_serialized.id == profile_scanner:
        print "Result is from Profile scan";
      elif result_serialized.id == session_scanner:
        print "Result is from Session scan";
      
      reader = libHyperPython.SerializedCellsReader(result_serialized.cells, len(result_serialized.cells))
      while reader.has_next():
        cell = reader.get_cell();
        print cell;

    client.async_scanner_close(profile_scanner);
    client.async_scanner_close(session_scanner);
    client.future_close(ff);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
  
  print "[async scanner - result as arrays]";

  try:
    ns = client.namespace_open("test");
    ff = client.future_open(0);

    row_intervals = [ RowInterval("1", True, "1", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss);
    
    row_intervals = [ RowInterval("0001-200238", True, "0001-200238", True) ];
    ss = ScanSpec(row_intervals = row_intervals);
    session_scanner = client.async_scanner_open(ns, "Session", ff, ss);
    
    while True:
      result_as_arrays = client.future_get_result_as_arrays(ff, 0);

      if result_as_arrays.is_empty:
        break;

      if result_as_arrays.is_error:
        print "Async scanner error:  {0}".format(result_as_arrays.error_msg);
        sys.exit(1);
      
      assert result_as_arrays.is_scan;
      assert result_as_arrays.id == profile_scanner or result_as_arrays.id == session_scanner;

      if result_as_arrays.id == profile_scanner:
        print "Result is from Profile scan";
      elif result_as_arrays.id == session_scanner:
        print "Result is from Session scan";
      
      for cell_as_array in result_as_arrays.cells:
        print_cell_as_array(cell_as_array);

    client.async_scanner_close(profile_scanner);
    client.async_scanner_close(session_scanner);
    client.future_close(ff);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

def test_atomic_counter(client):
  
  print "[atomic counter]";
  
  try:
    ns = client.namespace_open("test");

    command = "CREATE TABLE Hits (count COUNTER)";
    result = client.hql_query(ns, command);

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  print "[atomic counter - increment]";

  try:
    ns = client.namespace_open("test");

    cells = [];
    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "1");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "1");
    cells.append(cell);
    
    mutator = client.mutator_open(ns, "Hits", 0, 0);
    client.mutator_set_cells(mutator, cells);
    client.mutator_flush(mutator);
    
    scanner = client.scanner_open(ns, "Hits", ScanSpec());
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
    
    client.scanner_close(scanner);
    client.mutator_close(mutator);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);
    
  try:
    ns = client.namespace_open("test");

    cells = [];
    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "=0");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "7");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:18");
    cell = Cell(key, "-1");
    cells.append(cell);

    key = Key(row = "/index.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "-2");
    cells.append(cell);

    key = Key(row = "/foo/bar.html", column_family = "count", column_qualifier = "2014-06-14 07:31:19");
    cell = Cell(key, "=19");
    cells.append(cell);

    mutator = client.mutator_open(ns, "Hits", 0, 0);
    client.mutator_set_cells(mutator, cells);
    client.mutator_flush(mutator);
    
    scanner = client.scanner_open(ns, "Hits", ScanSpec());
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell;
      
    client.scanner_close(scanner);
    client.mutator_close(mutator);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message
    sys.exit(1);

def test_unique(client):
  
  print "[unique]";
  
  try:
    ns = client.namespace_open("test");

    command = "CREATE TABLE User (info, id TIME_ORDER desc MAX_VERSIONS 1)";
    result = client.hql_query(ns, command);

    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  try:
    ns = client.namespace_open("test");
    
    key = Key(row = "joe1987", column_family = "id");
    ret = client.create_cell_unique(ns, "User", key, "");
    
    key = Key(row = "mary.bellweather", column_family = "id");
    ret = client.create_cell_unique(ns, "User", key, "");
    
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

  try:
    ns = client.namespace_open("test");
    
    key = Key(row = "joe1987", column_family = "id");
    ret = client.create_cell_unique(ns, "User", key, "");
    
    client.namespace_close(ns);
    
  except ClientException as e:
    if e.code == 48:  # ALREADY_EXISTS
      print "User name '{0}' is already taken".format(key.row);
      client.namespace_close(ns);
    else:
      print e.message;
      sys.exit(1);

  try:
    ns = client.namespace_open("test");
    scanner = client.scanner_open(ns, "User", ScanSpec());
    
    while True:
      cells = client.scanner_get_cells(scanner);
      if not cells:
        break;
      for cell in cells:
        print cell.key;
    
    client.scanner_close(scanner);
    client.namespace_close(ns);
    
  except ClientException as e:
    print e.message;
    sys.exit(1);

try:

  try:
    client = ThriftClient("localhost", 15867)
  except:
    print sys.exc_info()
    sys.exit(1);

  test_basic(client);
  test_convenience(client);
  test_create_table(client);
  test_alter_table(client);
  test_mutator(client);
  test_scanner(client);
  test_hql(client);
  test_secondary_indices(client);
  test_async(client);
  test_atomic_counter(client);
  test_unique(client);
  
except:
  print sys.exc_info()
  raise

