#!/usr/bin/ruby
# -*- coding: utf-8 -*-
#
# Copyright (C) 2007-2014 Hypertable, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither Hypertable, Inc. nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
require 'rubygems'
require ENV['HYPERTABLE_HOME'] + '/lib/rb/hypertable/thrift_client'
require 'time'
include Hypertable::ThriftGen

def print_cell_as_array(cell_as_array)
  puts "{CellAsArray key={Key row=%s column_family=%s column_qualifier=%s} value=%s}" %
    [cell_as_array[0], cell_as_array[1], cell_as_array[2], cell_as_array[3]]
end

def test_basic(client)

  puts "[basic]"

  begin
    
    if !client.namespace_exists("test")
      client.namespace_create("test")
    end

    ns = client.namespace_open("test")

    if_exists = true
    client.table_drop(ns, "Fruits", if_exists)

    cf_specs = Hash.new

    cf = ColumnFamilySpec.new
    cf.name = "genus"
    cf_specs["genus"] = cf

    cf = ColumnFamilySpec.new
    cf.name = "description"
    cf_specs["description"] = cf

    cf = ColumnFamilySpec.new
    cf.name = "tag"
    cf_specs["tag"] = cf

    schema = Schema.new

    schema.column_families = cf_specs

    client.table_create(ns, "Fruits", schema)

    client.namespace_create("/test/sub")

    listing = client.namespace_get_listing(ns)

    listing.each do |entry|
      if entry.is_namespace
        puts "%s\t(dir)" % entry.name
      else
        puts entry.name
      end
    end

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

end

def test_convenience(client)

  puts "[set cells]"

  begin
    ns = client.namespace_open("test")

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "apple"
    cell.key.column_family = "genus"
    cell.value = "Malus"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "apple"
    cell.key.column_family = "description"
    cell.value = "The apple is the pomaceous fruit of the apple tree."
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "apple"
    cell.key.column_family = "tag"
    cell.key.column_qualifier = "crunchy"
    cells.push(cell)

    client.set_cells(ns, "Fruits", cells)

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[get cells]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new
    ss.columns = Array[ "description" ]

    cells = client.get_cells(ns, "Fruits", ss)

    cells.each { |cell| puts cell }

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[set cells as arrays]"

  begin
    ns = client.namespace_open("test")

    cells_as_arrays = Array.new

    cell_as_array = Array["orange", "genus", "", "Citrus"]
    cells_as_arrays.push(cell_as_array)

    cell_as_array = Array["orange", "description", "", "The orange (specifically," +
                          "the sweet orange) is the fruit of the citrus species" +
                          "Citrus × sinensis in the family Rutaceae.""Citrus"]
    cells_as_arrays.push(cell_as_array)

    cell_as_array = Array["orange", "tag", "juicy", ""]
    cells_as_arrays.push(cell_as_array)

    client.set_cells_as_arrays(ns, "Fruits", cells_as_arrays)

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[get cells as arrays]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new
    ss.columns = Array[ "description" ]

    cells_as_arrays = client.get_cells_as_arrays(ns, "Fruits", ss)

    cells_as_arrays.each { |cell_as_array| print_cell_as_array(cell_as_array) }

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_create_table(client)

  puts "[create table]"

  begin
    ns = client.namespace_open("test")

    schema = Schema.new

    schema.access_groups = Hash.new
    schema.column_families = Hash.new

    # Set table defaults
    schema.access_group_defaults = AccessGroupOptions.new
    schema.access_group_defaults.blocksize = 65536
    schema.column_family_defaults = ColumnFamilyOptions.new
    schema.column_family_defaults.max_versions = 1

    # Access group "ag_normal"
    ag_spec = AccessGroupSpec.new
    ag_spec.defaults = ColumnFamilyOptions.new
    ag_spec.defaults.max_versions = 2
    ag_spec.name = "ag_normal"
    schema.access_groups["ag_normal"] = ag_spec

    # Column "a"
    cf_spec = ColumnFamilySpec.new
    cf_spec.name = "a"
    cf_spec.access_group = "ag_normal"
    cf_spec.value_index = true
    cf_spec.qualifier_index = true
    schema.column_families["a"] = cf_spec

    # Column "b"
    cf_spec = ColumnFamilySpec.new
    cf_spec.options = ColumnFamilyOptions.new
    cf_spec.options.max_versions = 3
    cf_spec.name = "b"
    cf_spec.access_group = "ag_normal"
    schema.column_families["b"] = cf_spec

    # Access group "ag_fast"
    ag_spec = AccessGroupSpec.new
    ag_spec.options = AccessGroupOptions.new
    ag_spec.options.in_memory = true
    ag_spec.options.blocksize = 131072
    ag_spec.name = "ag_fast"
    schema.access_groups["ag_fast"] = ag_spec    

    # Column "c"
    cf_spec = ColumnFamilySpec.new
    cf_spec.name = "c"
    cf_spec.access_group = "ag_fast"
    schema.column_families["c"] = cf_spec

    # Access group "ag_secure"
    ag_spec = AccessGroupSpec.new
    ag_spec.options = AccessGroupOptions.new
    ag_spec.options.replication = 5
    ag_spec.name = "ag_secure"
    schema.access_groups["ag_secure"] = ag_spec    

    # Column "d"
    cf_spec = ColumnFamilySpec.new
    cf_spec.name = "d"
    cf_spec.access_group = "ag_secure"
    schema.column_families["d"] = cf_spec

    # Access group "ag_counter"
    ag_spec = AccessGroupSpec.new
    ag_spec.defaults = ColumnFamilyOptions.new
    ag_spec.defaults.counter = true
    ag_spec.defaults.max_versions = 0
    ag_spec.name = "ag_counter"
    schema.access_groups["ag_counter"] = ag_spec

    # Column "e"
    cf_spec = ColumnFamilySpec.new
    cf_spec.name = "e"
    cf_spec.access_group = "ag_counter"
    schema.column_families["e"] = cf_spec

    # Column "f"
    cf_spec = ColumnFamilySpec.new
    cf_spec.options = ColumnFamilyOptions.new
    cf_spec.options.counter = false
    cf_spec.name = "f"
    cf_spec.access_group = "ag_counter"
    schema.column_families["f"] = cf_spec

    client.table_create(ns, "TestTable", schema)

    result = client.hql_query(ns, "SHOW CREATE TABLE TestTable")

    if (!result.results.empty?)
      puts result.results[0]
    end

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_alter_table(client)

  puts "[alter table]"

  begin
    ns = client.namespace_open("test")

    schema = client.get_schema(ns, "TestTable")

    # Rename column "b" to "z"
    cf_spec = schema.column_families["b"]
    schema.column_families.delete("b")
    cf_spec.name = "z"
    schema.column_families["z"] = cf_spec
    
    # Add column "g"
    cf_spec = ColumnFamilySpec.new
    cf_spec.name = "g"
    cf_spec.access_group = "ag_counter"
    schema.column_families["g"] = cf_spec

    client.table_alter(ns, "TestTable", schema)

    result = client.hql_query(ns, "SHOW CREATE TABLE TestTable")

    if (!result.results.empty?)
      puts result.results[0]
    end

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_mutator(client)

  puts "[mutator]"

  begin
    ns = client.namespace_open("test")

    mutator = client.mutator_open(ns, "Fruits", 0, 0)

    # Auto-assigned timestamps

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "lemon"
    cell.key.column_family = "genus"
    cell.value = "Citrus"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "lemon"
    cell.key.column_family = "tag"
    cell.key.column_qualifier = "bitter"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "lemon"
    cell.key.column_family = "description"
    cell.value = "The lemon (Citrus × limon) is a small evergreen tree native to Asia."
    cells.push(cell)

    client.mutator_set_cells(mutator, cells)
    client.mutator_flush(mutator)

    # Explicitly-supplied timestamps

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "mango"
    cell.key.column_family = "genus"
    cell.key.timestamp = Time.parse("2014-06-06 16:27:15").to_i * 1000000000
    cell.value = "Mangifera"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "mango"
    cell.key.column_family = "tag"
    cell.key.column_qualifier = "sweet"
    cell.key.timestamp = Time.parse("2014-06-06 16:27:15").to_i * 1000000000
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "mango"
    cell.key.column_family = "description"
    cell.key.timestamp = Time.parse("2014-06-06 16:27:15").to_i * 1000000000
    cell.value = "Mango is one of the delicious seasonal fruits grown in the tropics."
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "mango"
    cell.key.column_family = "description"
    cell.key.timestamp = Time.parse("2014-06-06 16:27:16").to_i * 1000000000
    cell.value = "The mango is a juicy stone fruit belonging to the genus " +
      "Mangifera, consisting of numerous tropical fruiting trees, that are" +
      " cultivated mostly for edible fruits."
    cells.push(cell)

    client.mutator_set_cells(mutator, cells)
    client.mutator_flush(mutator)
  
    # Delete cells

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "apple"
    cell.key.flag = Hypertable::ThriftGen::KeyFlag::DELETE_ROW
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "mango"
    cell.key.column_family = "description"
    cell.key.timestamp = Time.parse("2014-06-06 16:27:15").to_i * 1000000000
    cell.key.flag = Hypertable::ThriftGen::KeyFlag::DELETE_CELL
    cells.push(cell)

    client.mutator_set_cells(mutator, cells)
    client.mutator_flush(mutator)
    client.mutator_close(mutator)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_scanner(client)

  puts "[scanner full table scan]"

  begin
    ns = client.namespace_open("test")

    scanner = client.scanner_open(ns, "Fruits", ScanSpec.new)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)      
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[scanner restricted scan]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    # Return row range [lemon..orange)
    ri = RowInterval.new
    ri.start_row = "lemon"
    ri.start_inclusive = true
    ri.end_row = "orange"
    ri.end_inclusive = false
    ss.row_intervals = Array[ ri ]

    # Return columns "genus", "tag:bitter", "tag:sweet"
    ss.columns = Array["genus", "tag:bitter", "tag:sweet"]

    # Return only most recent version of each cell
    ss.versions = 1

    scanner = client.scanner_open(ns, "Fruits", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)      
    end

    client.scanner_close(scanner)

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_hql(client)

  puts "[hql query]"

  begin
    ns = client.namespace_open("test")

    result = client.hql_query(ns, "GET LISTING")

    result.results.each { |str| puts str }

    result = client.hql_query(ns, "SELECT * FROM Fruits WHERE ROW = 'mango'")

    result.cells.each { |cell| puts cell }

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[hql_query_as_arrays]"

  begin
    ns = client.namespace_open("test")

    result_as_arrays =
      client.hql_query_as_arrays(ns, "SELECT * FROM Fruits WHERE ROW = 'lemon'")

    result_as_arrays.cells.each { |cell_as_array| print_cell_as_array(cell_as_array) }

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[hql_exec mutator]"

  begin
    ns = client.namespace_open("test")

    result = client.hql_exec(ns, "INSERT INTO Fruits VALUES ('strawberry', " +
                             "'genus', 'Fragaria'), ('strawberry', 'tag:fibrous', '')," +
                             " ('strawberry', 'description', 'The garden strawberry is" +
                             " a widely grown hybrid species of the genus Fragaria')",
                             true, false)

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "pineapple"
    cell.key.column_family = "genus"
    cell.value = "Ananas"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "pineapple"
    cell.key.column_family = "tag"
    cell.key.column_qualifier = "acidic"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "pineapple"
    cell.key.column_family = "description"
    cell.value = "The pineapple (Ananas comosus) is a tropical plant with " +
      "edible multiple fruit consisting of coalesced berries."
    cells.push(cell)

    client.mutator_set_cells(result.mutator, cells)
    client.mutator_flush(result.mutator)
    client.mutator_close(result.mutator)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[hql_exec scanner]"

  begin
    ns = client.namespace_open("test")

    result = client.hql_exec(ns, "SELECT * FROM Fruits", false, true)

    cells = client.scanner_get_cells(result.scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(result.scanner)      
    end

    client.scanner_close(result.scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_secondary_indices(client)

  puts "[secondary incides]"

  begin
    ns = client.namespace_open("test")

    command = "CREATE TABLE products (title, section, info, category," +
      "INDEX section, INDEX info, QUALIFIER INDEX info," +
      "QUALIFIER INDEX category)"
    client.hql_query(ns, command)

    command = "LOAD DATA INFILE 'indices_test_products.tsv' INTO TABLE products"
    client.hql_query(ns, command)

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products WHERE section = 'books']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "section"
    column_predicate.operation = ColumnPredicateOperation::EXACT_MATCH
    column_predicate.value = "books"
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products WHERE info:actor = 'Jack Nicholson']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "actor"
    column_predicate.operation = ColumnPredicateOperation::EXACT_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Jack Nicholson"
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =^ 'Addison-Wesley']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "publisher"
    column_predicate.operation = ColumnPredicateOperation::PREFIX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Addison-Wesley"
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title", "info:publisher" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =~ /^Addison-Wesley/]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "publisher"
    column_predicate.operation = ColumnPredicateOperation::REGEX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "^Addison-Wesley"
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title", "info:publisher" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products WHERE Exists(info:studio)]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "studio"
    column_predicate.operation = ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products WHERE Exists(category:/^\\/Movies/)]"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    column_predicate = ColumnPredicate.new

    column_predicate.column_family = "category"
    column_predicate.column_qualifier = "^/Movies"
    column_predicate.operation = ColumnPredicateOperation::QUALIFIER_REGEX_MATCH
    ss.column_predicates = Array[ column_predicate ]

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products WHERE " +
    "info:author =~ /^Stephen P/ OR info:publisher =^ 'Anchor']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new
    ss.column_predicates = Array.new

    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "author"
    column_predicate.operation = ColumnPredicateOperation::REGEX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "^Stephen P"
    ss.column_predicates.push(column_predicate)

    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "publisher"
    column_predicate.operation = ColumnPredicateOperation::PREFIX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Anchor"
    ss.column_predicates.push(column_predicate)

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products " +
    "WHERE info:author =~ /^Stephen [PK]/ AND info:publisher =^ 'Anchor']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new
    ss.column_predicates = Array.new

    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "author"
    column_predicate.operation = ColumnPredicateOperation::REGEX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "^Stephen [PK]"
    ss.column_predicates.push(column_predicate)

    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "publisher"
    column_predicate.operation = ColumnPredicateOperation::PREFIX_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Anchor"
    ss.column_predicates.push(column_predicate)

    ss.and_column_predicates = true

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products " +
    "WHERE ROW > 'B00002VWE0' AND info:actor = 'Jack Nicholson']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    # ROW > 'B00002VWE0'
    ri = RowInterval.new
    ri.start_row = "B00002VWE0"
    ri.start_inclusive = false
    ss.row_intervals = Array[ ri ]

    # info:actor = 'Jack Nicholson'
    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "actor"
    column_predicate.operation = ColumnPredicateOperation::EXACT_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Jack Nicholson"
    ss.column_predicates = Array[ column_predicate ]

    ss.and_column_predicates = true

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[secondary index - SELECT title FROM products " +
    "WHERE ROW =^ 'B' AND info:actor = 'Jack Nicholson']"

  begin
    ns = client.namespace_open("test")

    ss = ScanSpec.new

    # ROW =^ 'B'
    ri = RowInterval.new
    ri.start_row = "B"
    ri.start_inclusive = true
    ri.end_row = "C"
    ri.end_inclusive = false
    ss.row_intervals = Array[ ri ]

    # info:actor = 'Jack Nicholson'
    column_predicate = ColumnPredicate.new
    column_predicate.column_family = "info"
    column_predicate.column_qualifier = "actor"
    column_predicate.operation = ColumnPredicateOperation::EXACT_MATCH |
      ColumnPredicateOperation::QUALIFIER_EXACT_MATCH
    column_predicate.value = "Jack Nicholson"
    ss.column_predicates = Array[ column_predicate ]

    ss.and_column_predicates = true

    ss.columns = Array[ "title" ]

    scanner = client.scanner_open(ns, "products", ss)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_async(client)

  puts "[async]"

  begin
    ns = client.namespace_open("test")

    command = "CREATE TABLE Profile (info, last_access MAX_VERSIONS 1)"
    client.hql_query(ns, command)

    command = "CREATE TABLE Session (user_id, page_hit)"
    client.hql_query(ns, command)

    command = "INSERT INTO Profile VALUES ('1', 'info:name', 'Joe'), ('2', 'info:name', 'Sue')"
    client.hql_query(ns, command)

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[async mutator]"

  begin
    ns = client.namespace_open("test")
    ff = client.future_open(0)
    profile_mutator = client.async_mutator_open(ns, "Profile", ff, 0)
    session_mutator = client.async_mutator_open(ns, "Session", ff, 0)

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "1"
    cell.key.column_family = "last_access"
    cell.value = "2014-06-13 16:06:09"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "2"
    cell.key.column_family = "last_access"
    cell.value = "2014-06-13 16:06:10"
    cells.push(cell)

    client.async_mutator_set_cells(profile_mutator, cells)

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "0001-200238"
    cell.key.column_family = "user_id"
    cell.key.column_qualifier = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "0001-200238"
    cell.key.column_family = "page_hit"
    cell.value = "/index.html"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "0002-383049"
    cell.key.column_family = "user_id"
    cell.key.column_qualifier = "2"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "0002-383049"
    cell.key.column_family = "page_hit"
    cell.value = "/foo/bar.html"
    cells.push(cell)
    client.async_mutator_set_cells(session_mutator, cells)

    client.async_mutator_flush(profile_mutator)
    client.async_mutator_flush(session_mutator)

    result_count = 0
    while true do
      result = client.future_get_result(ff, 0)
      if result.is_empty
        break
      end
      result_count += 1
      if result.is_error
        puts "Async mutator error:  " + result.error_msg
        exit 1
      end
      if result.id == profile_mutator
        puts "Result is from Profile mutation"
      elsif result.id == session_mutator
        puts "Result is from Session mutation"
      end
    end

    puts "result count = %d" % result_count

    client.async_mutator_close(profile_mutator)
    client.async_mutator_close(session_mutator)
    client.future_close(ff)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[async scanner]"

  begin
    ns = client.namespace_open("test")
    ff = client.future_open(0)

    ss = ScanSpec.new
    ri = RowInterval.new
    ri.start_row = "1"
    ri.start_inclusive = true
    ri.end_row = "1"
    ri.end_inclusive = true
    ss.row_intervals = Array[ ri ]
    profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss)

    ss = ScanSpec.new
    ri = RowInterval.new
    ri.start_row = "0001-200238"
    ri.start_inclusive = true
    ri.end_row = "0001-200238"
    ri.end_inclusive = true
    ss.row_intervals = Array[ ri ]
    session_scanner = client.async_scanner_open(ns, "Session", ff, ss)

    while true do
      result = client.future_get_result(ff, 0)
      if result.is_empty
        break
      end
      if result.is_error
        puts "Async mutator error:  " + result.error_msg
        exit 1
      end
      if result.id == profile_scanner
        puts "Result is from Profile scan"
      elsif result.id == session_scanner
        puts "Result is from Session scan"
      end
      result.cells.each { |cell| puts cell }
    end

    client.async_scanner_close(profile_scanner)
    client.async_scanner_close(session_scanner)
    client.future_close(ff)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[async scanner - result as arrays]"

  begin
    ns = client.namespace_open("test")
    ff = client.future_open(0)

    ss = ScanSpec.new
    ri = RowInterval.new
    ri.start_row = "1"
    ri.start_inclusive = true
    ri.end_row = "1"
    ri.end_inclusive = true
    ss.row_intervals = Array[ ri ]
    profile_scanner = client.async_scanner_open(ns, "Profile", ff, ss)

    ss = ScanSpec.new
    ri = RowInterval.new
    ri.start_row = "0001-200238"
    ri.start_inclusive = true
    ri.end_row = "0001-200238"
    ri.end_inclusive = true
    ss.row_intervals = Array[ ri ]
    session_scanner = client.async_scanner_open(ns, "Session", ff, ss)

    while true do
      result_as_arrays = client.future_get_result_as_arrays(ff, 0)
      if result_as_arrays.is_empty
        break
      end
      if result_as_arrays.is_error
        puts "Async mutator error:  " + result_as_arrays.error_msg
        exit 1
      end
      if result_as_arrays.id == profile_scanner
        puts "Result is from Profile scan"
      elsif result_as_arrays.id == session_scanner
        puts "Result is from Session scan"
      end
      result_as_arrays.cells.each { |cell_as_array| print_cell_as_array(cell_as_array) }
    end

    client.async_scanner_close(profile_scanner)
    client.async_scanner_close(session_scanner)
    client.future_close(ff)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_atomic_counters(client)

  puts "[atomic counter]"

  begin
    ns = client.namespace_open("test")
    result = client.hql_query(ns, "CREATE TABLE Hits (count COUNTER)")
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[atomic counter - increment]"

  begin
    ns = client.namespace_open("test")

    mutator = client.mutator_open(ns, "Hits", 0, 0)

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "1"
    cells.push(cell)

    client.mutator_set_cells(mutator, cells)
    client.mutator_flush(mutator)

    scanner = client.scanner_open(ns, "Hits", ScanSpec.new)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.mutator_close(mutator)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  puts "[atomic counter - reset and subtraction]"

  begin
    ns = client.namespace_open("test")

    mutator = client.mutator_open(ns, "Hits", 0, 0)

    cells = Array.new

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "=0"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "7"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:18"
    cell.value = "-1"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/index.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "-2"
    cells.push(cell)

    cell = Cell.new
    cell.key = Key.new
    cell.key.row = "/foo/bar.html"
    cell.key.column_family = "count"
    cell.key.column_qualifier = "2014-06-14 07:31:19"
    cell.value = "=19"
    cells.push(cell)

    client.mutator_set_cells(mutator, cells)
    client.mutator_flush(mutator)

    scanner = client.scanner_open(ns, "Hits", ScanSpec.new)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.mutator_close(mutator)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end

def test_unique(client)

  puts "[unique]"

  begin
    ns = client.namespace_open("test")
    client.hql_query(ns, "CREATE TABLE User (info, id TIME_ORDER desc MAX_VERSIONS 1)")
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  begin
    ns = client.namespace_open("test")

    key = Key.new
    key.column_family = "id"
    key.row = "joe1987"
    client.create_cell_unique(ns, "User", key, "")

    key = Key.new
    key.column_family = "id"
    key.row = "mary.bellweather"
    client.create_cell_unique(ns, "User", key, "")

    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end

  begin
    ns = client.namespace_open("test")

    key = Key.new
    key.column_family = "id"
    key.row = "joe1987"
    client.create_cell_unique(ns, "User", key, "")

    client.namespace_close(ns)

  rescue ClientException => e
    if e.code == 48
      puts "User name '%s' is already taken" % key.row
      client.namespace_close(ns)
    else
      puts "exception caught on line %d: %s" % [__LINE__, e.message]
      exit 1
    end
  end

  begin
    ns = client.namespace_open("test")

    scanner = client.scanner_open(ns, "User", ScanSpec.new)

    cells = client.scanner_get_cells(scanner)
    while !cells.empty? do
      cells.each { |cell| puts cell }
      cells = client.scanner_get_cells(scanner)
    end

    client.scanner_close(scanner)
    client.namespace_close(ns)

  rescue ClientException => e
    puts "exception caught on line %d: %s" % [__LINE__, e.message]
    exit 1
  end
end


begin

  Hypertable.with_thrift_client("localhost", 15867) do |client|
    test_basic(client)
    test_convenience(client)
    test_create_table(client)
    test_alter_table(client)
    test_mutator(client)
    test_scanner(client)
    test_hql(client)
    test_secondary_indices(client)
    test_async(client)
    test_atomic_counters(client)
    test_unique(client)
  end

rescue
  puts $!
  exit 1
end
